#include	<iostream>
#include	<iomanip>
#include	<fstream>
#include	<sstream>
#include	<string>
#include	<cmath>

using namespace std;

#include	<mlib/types.h>
#include	<mlib/array.h>
#include	"lum.h"
#include	"rgba.h"
#include	"smpte.h"
#include	"pnm.h"
#include	"pgm.h"

/////////////////////////// protected helper functions (for constructors) //////
void Pgm::init(int row, int col, ushort clr)
{
  init(row,col);
  if(!clr) {
    testPattern();
    rename("./","smpte.pgm");
  }
}

void Pgm::init(int row, int col)
{
  r		= row;
  c		= col;
  format	= PGM_FORMAT;
  maxval	= 255;
  fpath		= NULL;
  fname		= NULL;
  larr		= new Array<ushort>(r*c);
  darr		= NULL;

}

void Pgm::init(int row, int col, Array<ushort> array)
{
  r		= row;
  c		= col;
  format	= PGM_FORMAT;
  maxval	= 255;
  fpath		= NULL;
  fname		= NULL;
  larr		= new Array<ushort>(array);
  darr		= NULL;
}

void Pgm::init(const Pgm& rhs)
{
  r		= rhs.r;
  c		= rhs.c;
  format	= rhs.format;
  maxval	= rhs.maxval;
  fpath		= rhs.fpath ? strdup(rhs.fpath) : NULL;
  fname		= rhs.fname ? strdup(rhs.fname) : NULL;
  larr		= NULL;
  darr		= NULL;

  rename(rhs.fpath,rhs.fname);
  larrcpy(rhs.larr);
  darrcpy(rhs.darr);
}

Pgm* Pgm::subimg(int lft,int rgt,int bot,int top)
{
	int	nr = bot-top+1;
	int	nc = rgt-lft+1;
	Pgm	*simg = NULL;

// BUG:  The new simg DOESN'T HAVE A PROPER HEADER (maxval is missing)
//       (Need to get this from (*this) somehow.)

  // error checking:
  //   assert(0 <= top < bot < r)
  //   assert(0 <= lft < rgt < c)
  if( (0 <= top) && (top < bot) && (bot < r) ) {
    if( (0 <= lft) && (lft < rgt) && (rgt < c) ) {
      // create new image
      simg = new Pgm(nr,nc,(ushort)0);
      // crop out subimage, writing to subimage array starting at (0,0)
      for(int i=top, si=0; i<bot; si++, i++) {
        for(int j=lft, sj=0; j<rgt; sj++, j++) {
          (*simg)(si,sj) = (*this)(i,j);
        }
      }
      return simg;
    }
  }
  return NULL;
}

double Pgm::mse(const Pgm& rhs)
{
	double	diff,mserror=0.0;
	double	rc = r*c;

  // calculate the mean squared error
  for(int i=0;i<r;i++) {
    for(int j=0;j<c;j++) {
      diff = fabs((double)((*this)(i,j) - rhs(i,j)));	// error
      mserror += diff*diff;				// squared error
    }
  }
  mserror /= rc;					// mean squared error
  return(mserror);
}

double Pgm::rms(const Pgm& rhs)
{
  // calculate the root mean squared error (square root of mse)
  return(sqrt(mse(rhs)));
}

double Pgm::snr(const Pgm& rhs)
{
  // calculate the (mean squared) signal-to-noise ratio (in decibels, dB)
  // assuming rhs is the reference (original) image
  // snr is defined as signal/noise, where
  //   signal = (*this) = g(i,j) = f(i,j) + e(i,j)
  //   error = e(i,j) = g(i,j) - f(i,j)
  // where f(i,j) is the reference image.

	double	dval,dmean=0.0,diff,mserror=0.0,ratio;

  for(int i=0;i<r;i++) {
    for(int j=0;j<c;j++) {
//    dval = (double)(rhs(i,j));			// reference image mean
      dval = (double)((*this)(i,j));			// "signal" image mean
      dmean += dval*dval;				// mean squard
      diff = fabs((double)((*this)(i,j) - rhs(i,j)));	// error
      mserror += diff*diff;				// squared error
    }
  }
  ratio = 10.0*log10(dmean/mserror);			// snr (ms)
  return(ratio);
}

/////////////////////////// operators: (const Pgm&, const Pgm&) ////////////////
const Pgm& Pgm::operator=(const Pgm& rhs)
{
  return copy(rhs);
}

const Pgm& Pgm::copy(const Pgm& rhs)
{
  if(this != &rhs) {
    if( (r != rhs.r) || (c != rhs.c) ) {
      delete larr;  delete darr;
      r = rhs.r;  c = rhs.c;
    }
    larrcpy(rhs.larr);
    darrcpy(rhs.darr);
  }

  return *this;
}

/////////////////////////// friends ////////////////////////////////////////////
ostream& operator<<(ostream& s,Pgm& rhs)
{
  rhs.putHeader(s);
  rhs.putImage_btot(s);
  return s;
}

istream& operator>>(istream& s,Pgm& rhs)
{
  rhs.getHeader(s);
  rhs.getImage_btot(s);
  return s;
}

/////////////////////////// public member functions ////////////////////////////
void Pgm::read(const char* fp,const char* fn)
{
	ostringstream	file;

  rename(fp,fn);

  file << fp << '/' << fn << '\0';

  ifstream ifs(file.str().c_str());
  if(!ifs) {
    cerr << "Can't open input " << file.str() << endl;
    exit(1);
  }

  getHeader(ifs);
  getImage(ifs);

  ifs.close();
}

void Pgm::write(const char* fp,const char* fn)
{
	ostringstream	file;

  rename(fp,fn);

  file << fp << '/' << fn << '\0';

  ofstream ofs(file.str().c_str());
  if(!ofs) {
    cerr << "Can't open output " << file.str() << endl;
    exit(1);
  }

  putHeader(ofs);
  putImage(ofs);

  ofs.close();
}

void Pgm::resize(int nr, int nc)
{
  if( (nr != r) || (nc != c) ) {
    ltod();
    darr = new Array<double>(*dresize(darr,nr,nc));
    r = nr; c = nc;
    dtol();
  }
}

void Pgm::rotate(double angle,int bilinear)
{
  ltod();
  darrrot(darr,angle,bilinear);
  dtol();
}

void Pgm::thresh(double min,double max)
{
  ltod();
  darrthresh(darr,min,max);
  dtol();
}

void Pgm::inverse(void)
{
	register uchar	grey;
	register uchar	alpha=0xff;

  for(int i=0;i<r;i++) {
    for(int j=0;j<c;j++) {
      grey		= (uchar)SHORT_L((*this)(i,j));
      grey		= (uchar)maxval - grey;
      (*this)(i,j)	= ((grey<<8) | alpha);
    }
  }
}

/////////////////////////// protected member functions /////////////////////////
void Pgm::getHeader(istream& s)
{
	int	ich1, ich2;

  ich1 = s.get();
  if(s.eof()) {
    cerr << "getHeader: EOF / read error reading magic number\n";
    exit(1);
  }
  ich2 = s.get();
  if(s.eof()) {
    cerr << "getHeader: EOF / read error reading magic number\n";
    exit(1);
  }
  format = ich1 * 256 + ich2;

  switch(PGM_FORMAT_TYPE(format)) {
    case PGM_TYPE:
      c = getInt(s);
      r = getInt(s);
      maxval = getInt(s);
      if(maxval > PGM_MAXMAXVAL) {
        cerr << "getHeader: maxval is too large - ";
        cerr << "try reconfiguring with PGM_BIGGRAYS\n";
        exit(1);
      }
      break;
    case PBM_TYPE:
      c = getInt(s);
      r = getInt(s);
      maxval = 1;
      break;
    default:
      cerr << "getHeader: bad magic number - not a pgm or pbm file\n";
      exit(1);
  }
}

void Pgm::putHeader(ostream& s)
{
  s << PGM_MAGIC1 << RPGM_MAGIC2 << endl;
  s << c << " " << r << endl;
  s << (int)maxval << endl;
}

void Pgm::getImage(istream& s)
{
	int		i,j;
	register int	bitshift;
	register gray	g, item;
	register bit	b;
	register uchar	grey;
	register uchar	alpha=0xff;

  if(larr) delete larr;
  larr = new Array<ushort>(r*c);

  for(i=0;i<r;i++) {
    bitshift = -1;
    for(j=0;j<c;j++) {
      switch(format) {
        case PGM_FORMAT:
          g = getInt(s);
          grey = (uchar)g;
          break;
        case RPGM_FORMAT:
          g = getRawByte(s);
          grey = (uchar)g;
          break;
        case PBM_FORMAT:
        case RPBM_FORMAT:
          switch(format) {
            case PBM_FORMAT:
              b = getBit(s);
              break;
            case RPBM_FORMAT:
              if(bitshift == -1) {
                item = getRawByte(s);
                bitshift = 7;
              }
              b = (item >> bitshift) & 1;
              --bitshift;
              break;
            default:
              cerr << "getImage: unknown PBM format\n";
              exit(1);
          }
          g = (b==PBM_WHITE) ? maxval : 0;
          grey = (uchar)g;
          break;
        default:
          cerr << "getImage: unknown PGM format\n";
          exit(1);
      }
#ifdef ANM_OSX
      // store in GL_LUMINANCE format (in a 16-bit unsigned integer)
      (*this)(i,j) = ((grey<<8) | alpha);
#else
      // store in GL_LUMINANCE format (in a 16-bit unsigned integer)
      (*this)(i,j) = ((alpha<<8) | grey);
#endif
    }
  }
}

void Pgm::getImage_btot(istream& s)
{
	int		i,j;
	register int	bitshift;
	register gray	g, item;
	register bit	b;
	register uchar	grey;
	register uchar	alpha=0xff;

  if(larr) delete larr;
  larr = new Array<ushort>(r*c);

  for(i=r-1;i>=0;i--) {
    bitshift = -1;
    for(j=0;j<c;j++) {
      switch(format) {
        case PGM_FORMAT:
          g = getInt(s);
          grey = (uchar)g;
          break;
        case RPGM_FORMAT:
          g = getRawByte(s);
          grey = (uchar)g;
          break;
        case PBM_FORMAT:
        case RPBM_FORMAT:
          switch(format) {
            case PBM_FORMAT:
              b = getBit(s);
              break;
            case RPBM_FORMAT:
              if(bitshift == -1) {
                item = getRawByte(s);
                bitshift = 7;
              }
              b = (item >> bitshift) & 1;
              --bitshift;
              break;
            default:
              cerr << "getImage: unknown PBM format\n";
              exit(1);
          }
          g = (b==PBM_WHITE) ? maxval : 0;
          grey = (uchar)g;
          break;
        default:
          cerr << "getImage: unknown PGM format\n";
          exit(1);
      }
#ifdef ANM_OSX
      // store in GL_LUMINANCE format (in a 16-bit unsigned integer)
      (*this)(i,j) = ((grey<<8) | alpha);
#else
      // store in GL_LUMINANCE format (in a 16-bit unsigned integer)
      (*this)(i,j) = ((alpha<<8) | grey);
#endif
    }
  }
}

void Pgm::putImage(ostream& s)
{
	int		i,j,charcount;
	register gray	g;
	register uchar	grey;

#ifdef PBMPLUS_RAWBITS
  for(i=0;i<r;i++) {
    for(j=0;j<c;j++) {
      grey	= (uchar)SHORT_L((*this)(i,j));
      g		= (gray)grey;
      s.put(g);
    }
  }
#else
  for(i=0;i<r;i++) {
    charcount = 0;
    for(j=0;j<c;j++) {
      grey	= (uchar)SHORT_L((*this)(i,j));
      g		= (gray)grey;
      if(charcount >= 65) {
        s.put('\n');
        charcount = 0;
      }
      else if(charcount > 0) {
        s.put(' ');
        charcount++;
      }
      putUS((ushort)g,s);
      charcount += 3;
    }
    if(charcount > 0)
      s.put('\n');
  }
#endif
}

void Pgm::putImage_btot(ostream& s)
{
	int		i,j,charcount;
	register gray	g;
	register uchar	grey;

#ifdef PBMPLUS_RAWBITS
  for(i=r-1;i>=0;i--) {
    for(j=0;j<c;j++) {
      grey	= (uchar)SHORT_L((*this)(i,j));
      g		= (gray)grey;
      s.put(g);
    }
  }
#else
  for(i=r-1;i>=0;i--) {
    charcount = 0;
    for(j=0;j<c;j++) {
      grey	= (uchar)SHORT_L((*this)(i,j));
      g		= (gray)grey;
      if(charcount >= 65) {
        s.put('\n');
        charcount = 0;
      }
      else if(charcount > 0) {
        s.put(' ');
        charcount++;
      }
      putUS((ushort)g,s);
      charcount += 3;
    }
    if(charcount > 0)
      s.put('\n');
  }
#endif
}

void Pgm::testPattern(void)
{
	int	tw=c/7+1,th=2*r/3+1;	// top bars: width, height
	int	bw=c/5+1,bh=1*r/3+1;	// bot bars: width, height

  for(int i=0;i<r;i++) {
    for(int j=0;j<c;j++) {
      if(i <= th) {		// top bars
        switch(j/tw) {
          case SMPTE_TOP_WHITE:   (*larr)[(i*c+j)] = Lum(SMPTE_WHITE_32);
            break;
          case SMPTE_TOP_YELLOW:  (*larr)[(i*c+j)] = Lum(SMPTE_YELLOW_32);
            break;
          case SMPTE_TOP_CYAN:    (*larr)[(i*c+j)] = Lum(SMPTE_CYAN_32);
            break;
          case SMPTE_TOP_GREEN:   (*larr)[(i*c+j)] = Lum(SMPTE_GREEN_32);
            break;
          case SMPTE_TOP_MAGENTA: (*larr)[(i*c+j)] = Lum(SMPTE_MAGENTA_32);
            break;
          case SMPTE_TOP_RED:     (*larr)[(i*c+j)] = Lum(SMPTE_RED_32);
            break;
          case SMPTE_TOP_BLUE:    (*larr)[(i*c+j)] = Lum(SMPTE_BLUE_32);
            break;
        }
      }
      else {			// bot bars
        switch(j/bw) {
          case SMPTE_BOT_BLUE1:   (*larr)[(i*c+j)] = Lum(SMPTE_BLUE_32);
            break;
          case SMPTE_BOT_WHITE:   (*larr)[(i*c+j)] = Lum(SMPTE_100WHITE_32);
            break;
          case SMPTE_BOT_BLUE2:   (*larr)[(i*c+j)] = Lum(SMPTE_BLUE_32);
            break;
          case SMPTE_BOT_BLACK1:  (*larr)[(i*c+j)] = Lum(SMPTE_BLACK_32);
            break;
          case SMPTE_BOT_BLACK2:  (*larr)[(i*c+j)] = Lum(SMPTE_BLACK_32);
            break;
        }
      }
    }
  }
}

ushort Pgm::Lum(ulong val)
{
	register uchar	red,green,blue;
	register uchar	alpha=0xff;
	register uchar	grey;
	ushort		g;

  red       = (uchar)LONG_R(val);
  green     = (uchar)LONG_G(val);
  blue      = (uchar)LONG_B(val);
  grey      = (gray)ILUM(red,green,blue);
  g = ((grey<<8) | alpha);
  return(g);
}

void Pgm::larrshift(int x,int y)
{
        Array<ushort>    *tlarr = new Array<ushort>((*larr));

  for(int i=0;i<r;i++) {
    for(int j=0;j<c;j++) {
      (*larr)[M2A(i,j,r,c)] = (*tlarr)[M2A(i+y,j+x,r,c)];
    }
  }
  delete tlarr;
}

/////////////////////////// private member functions ///////////////////////////
void Pgm::ltod(void)
{
	double		dr,dg,db;
	register uchar	grey;

  if(!larr) return;

  if(darr) delete darr;
  darr = new Array<double>(r*c);

  for(int i=0;i<r;i++) {
    for(int j=0;j<c;j++) {
      grey      = (uchar)SHORT_L((*this)(i,j));
      (*darr)[M2A(i,j,r,c)] = (double)grey;
    }
  }
}

void Pgm::dtol(void)
{
	register uchar	alpha=0xff;
	register uchar	grey;

  if(!darr) return;

  if(larr) delete larr;
  larr = new Array<ushort>(r*c);

  for(int i=0;i<r;i++) {
    for(int j=0;j<c;j++) {
      grey = static_cast<uchar>((*darr)[M2A(i,j,r,c)]);
      (*this)(i,j) = ((grey<<8) | alpha);
    }
  }
}
