#ifndef PGM_H
#define PGM_H

class Pgm : public Pnm {
	public:
	// constructors (overloaded)
	Pgm(int row=256,int col=256,ushort clr=0)	{init(row,col,clr);}
	Pgm(int row=256,int col=256)			{init(row,col);}
	Pgm(Array<ushort> array, int row, int col)	{init(row,col,array);}
	Pgm(const Pgm& rhs)				{init(rhs);}
	Pgm(Pgm* rhs)					{init(*rhs);}

	// destructors
virtual	~Pgm()
	{
	  if(larr)	delete larr;
	  if(darr)	delete darr;
	  if(fpath)	delete [] fpath;
	  if(fname)	delete [] fname;
	}

	// operators: dereferencing (Fortran-like indeces), with wraparound
	ushort& operator()(int i,int j)
	  { return((*larr)[M2A(i,j,r,c)]); }
	ushort& operator()(int i,int j) const
	  { return((*larr)[M2A(i,j,r,c)]); }

	// operators: (const Pgm&, const Pgm&)
	const Pgm& operator=(const Pgm& rhs);

	// operators: (const Pgm&, Pgm*)
	const Pgm& operator=(Pgm* rhs)
	  { return((*this) = (*rhs)); }

	// public helper function
	const Pgm& copy(const Pgm& rhs);

	// friends
friend	ostream&	operator<<(ostream& s, Pgm& rhs);
friend	ostream&	operator<<(ostream& s, Pgm* rhs)
	  { return(s << (*rhs)); }
friend	istream&	operator>>(istream& s, Pgm& rhs);
friend	istream&	operator>>(istream& s, Pgm* rhs)
	  { return(s >> (*rhs)); }

	// public member functions
virtual	void	read(const char* fp, const char* fn);
virtual	void	write(const char* fp, const char* fn);
virtual	void	resize(int nr, int nc);
virtual	void	rotate(double angle,int bilinear);
virtual void	thresh(double min,double max);
	void	inverse(void);
	ushort*	data()
	  { return(larr->data()); }

	Pgm*	subimg(int,int,int,int);
	double	mse(const Pgm& rhs);
	double	mse(Pgm* rhs)			{ return(mse(*rhs)); }
	double	rms(const Pgm& rhs);
	double	rms(Pgm* rhs)			{ return(rms(*rhs)); }
	double	snr(const Pgm& rhs);
	double	snr(Pgm* rhs)			{ return(snr(*rhs)); }

	// protected: available to this (base) class and subs. derived classes
	protected:
	void	init(int row,int col);
	void	init(int row,int col,ushort clr);
	void	init(int row,int col,Array<ushort> array);
	void	init(const Pgm& rhs);
virtual	void	getHeader(istream& s);
virtual	void	putHeader(ostream& s);
virtual	void	getImage(istream& s);
virtual	void	getImage_btot(istream& s);
virtual	void	putImage(ostream& s);
virtual	void	putImage_btot(ostream& s);
	void	testPattern(void);
	ushort	Lum(ulong val);

	void	larrcpy(Array<ushort>* slarr)
	  { if(slarr) larr = new Array<ushort>((*slarr)); }
	void	larrshift(int x,int y);

	// private: (was private, but I made it protected so that PgmCell
	// can inherit these.
	void	dtol(void);
	void	ltod(void);
	void	darrcpy(Array<double>* srcdarr)
	  { if(srcdarr) darr = new Array<double>((*srcdarr)); }

	Array<ushort>*	larr;
	Array<double>*	darr;

	// private: only available to this class
	private:
};

#endif
