#include	<iostream>
#include	<iomanip>
#include	<fstream>
#include	<sstream>
#include	<string>
#include	<cmath>

using namespace std;

#include	<mlib/types.h>
#include	<mlib/array.h>
#include	<mlib/matrix.h>
#include	<mlib/vector.h>
#include	"rgba.h"
#include	"smpte.h"
#include	"pnm.h"

/////////////////////////// protected member functions /////////////////////////
int Pnm::getInt(istream& s)
{
	register char 	ch;
	register int 	i=0;

  do {
    ch = getChar(s);
  } while(ch == ' ' || ch == '\t' || ch == '\n' || ch == '\r');

  if(ch < '0' || ch > '9') {
    cerr << "getInt: junk in file where an integer should be\n";
    exit(1);
  }
  do {
    i = i * 10 + ch - '0';
    ch = getChar(s);
  } while (ch >= '0' && ch <= '9');

  return i;
}

char Pnm::getChar(istream& s)
{
	register int 	ich;
	register char 	ch;

  ich = s.get();
  if(s.eof()) {
    cerr << "getChar: EOF / read error\n";
    exit(1);
  }
  ch = (char)ich;

  if(ch == '#') {
    do {
      ich = s.get();
      if(s.eof()) {
        cerr << "getChar: EOF / read error\n";
        exit(1);
      }
      ch = (char)ich;
    } while (ch != '\n' && ch != '\r');
  }
  return ch;
}

byte Pnm::getRawByte(istream& s)
{
	register int 	iby;

  iby = s.get();
  if(s.eof()) {
    cerr << "getRawByte: EOF / read error\n";
    exit(1);
  }
  return (uchar) iby;
}

bit Pnm::getBit(istream& s)
{
	register char 	ch;

  do {
    ch = getChar(s);
  }
  while(ch == ' ' || ch == '\t' || ch == '\n' || ch == '\r');

  if(ch != '0' && ch != '1') {
    cerr << "getBit: junk in file where bits should be\n";
  }
  return (ch == '1') ? 1 : 0;
}

void Pnm::putUS(ushort n, ostream& s)
{
  if(n >= 10)
    putUS(n/10,s);
  s.put(n%10 + '0');
}

void Pnm::fupdate(char*& fdest, const char* fsrc)
{
  if(fdest) delete [] fdest;
  fdest = new char [strlen(fsrc) + 1];
  strcpy(fdest,fsrc);
}

void Pnm::darrrot(Array<double>* darr,double angle,int bilinear)
{
	int		ri,rj;
	double		cx = c/2., cy = r/2.;
	double		ex,ey;
	double		theta = angle/180.*M_PI;
	Matrix<double>	T1(3,3),T2(3,3),R(3,3),M(3,3);
	Vector<double>	v(0., 0., 0.), vo(0., 0., 0.);
	Array<double>	*tdarr = new Array<double>((*darr));

  // set up transformation matrix, including the translation to
  // origin (center of image)
  T1(0,0) =         1.0;  T1(0,1) =         0.0;  T1(0,2) = 0.0;
  T1(1,0) =         0.0;  T1(1,1) =         1.0;  T1(1,2) = 0.0;
  T1(2,0) =         -cx;  T1(2,1) =         -cy;  T1(2,2) = 1.0;

  // use -ve version of rotation matrix (e.g., CW rotation) for inverse
  // mapping, so end result will be CCW rotation
  R(0,0) =  cos(theta);  R(0,1) =  sin(theta);  R(0,2) = 0.0;
  R(1,0) = -sin(theta);  R(1,1) =  cos(theta);  R(1,2) = 0.0;
  R(2,0) =         0.0;  R(2,1) =         0.0;  R(2,2) = 1.0;

  T2(0,0) =         1.0;  T2(0,1) =         0.0;  T2(0,2) = 0.0;
  T2(1,0) =         0.0;  T2(1,1) =         1.0;  T2(1,2) = 0.0;
  T2(2,0) =          cx;  T2(2,1) =          cy;  T2(2,2) = 1.0;

  // overall transformation, concatenated into 1 3x3 homogeneous coordinate
  // matrix
  M  = T1 * R;
  M *= T2;

  // clear out destination image
  darr->black();

  // Perform the transformation:  do inverse mapping.  Note that we'd
  // like to rotate the source coordinates (i,j) to obtain the destination
  // coordinates (ri,rj), but doing so generates holes in the resultant image
  // (due to round-off error).  So instead do an inverse map where we rotate
  // the destination coordinates (i,j), and obtain the pixel info from the
  // source image at location (ri,rj).  This generates a full image, albeit
  // still a little imperfect, this time probably due to undersampling
  // of the original.  To fix this, calculate the round-off error, then
  // obtain averages via bilinear interpolation (which gives a weighted
  // average):
  //   dx = x - floor(x)  // error in x coordinate
  //   dy = y - floor(y)  // error in y coordinate
  //   x = (1 - dy)(1 - dx)x[j  ,i  ] +
  //       (1 - dy)(1 - dx)x[j+1,i  ] +
  //       (1 - dy)(1 - dx)x[j  ,i+1] +
  //       (1 - dy)(1 - dx)x[j+1,i+1]
  //
  //                x[j  ,i  ]     x[j+1,i  ]
  //                     +--------------+
  //                     |      .       |
  //                     |      . dy    |
  //                     |      .       |
  //                     |. . . x       |
  //                     |  dx          |
  //                     |              |
  //                     +--------------+
  //                x[j  ,i+1]     x[j+1,i+1]
  //

  for(int i=0;i<r;i++) {
    for(int j=0;j<c;j++) {
      vo[0] = j;  vo[1] = i;  vo[2] = 1.0;
      v = vo * M;
      rj = (int)v[0];  ri = (int)v[1];
      if( ((0 <= ri) && (ri < r)) && ((0 <= rj) && (rj < c)) ) {
        if(bilinear) {
          // calculate round-off error (gives the weights for interpolation)
          ex = v[0] - floor(v[0]);  ey = v[1] - floor(v[1]);
          // interpolate
          (*darr)[M2A(i,j,r,c)] =
            (1.0-ey)*(1.0 - ex) * (*tdarr)[M2A(ri  ,rj  ,r,c)] + 
            (1.0-ey)*(      ex) * (*tdarr)[M2A(ri  ,rj+1,r,c)] + 
            (    ey)*(1.0 - ex) * (*tdarr)[M2A(ri+1,rj  ,r,c)] + 
            (    ey)*(      ex) * (*tdarr)[M2A(ri+1,rj+1,r,c)];
        }
        else {
          // no interpolation
          (*darr)[M2A(i,j,r,c)] = (*tdarr)[M2A(ri,rj,r,c)];
        }
      }
    }
  }
  delete tdarr;
}

void Pnm::darrthresh(Array<double>* darr,double min,double max)
{
	Array<double>	*tdarr;

  tdarr = new Array<double>((*darr));

  for(int i=0;i<r;i++) {
    for(int j=0;j<c;j++) {
    //
    // this thresholding turns pixels outside the range to black
    //
    //if( (min <= (*tdarr)[M2A(i,j,r,c)]) && ((*tdarr)[M2A(i,j,r,c)] <= max) )
    //  (*darr)[M2A(i,j,r,c)] = (*tdarr)[M2A(i,j,r,c)];
    //else
    //  (*darr)[M2A(i,j,r,c)] = 0.0;
    //
    // this thresholding turns pixels < min to black, > max to white
    //
      if( (*tdarr)[M2A(i,j,r,c)] < min )
        (*darr)[M2A(i,j,r,c)] = 0.0;
      else
      if( (*tdarr)[M2A(i,j,r,c)] > max )
        (*darr)[M2A(i,j,r,c)] = (double)maxval;
      else
        (*darr)[M2A(i,j,r,c)] = (*tdarr)[M2A(i,j,r,c)];
    }
  }
  delete tdarr;
}

void Pnm::darrshift(Array<double>* darr,int x,int y)
{
	Array<double>	*tdarr;

  tdarr = new Array<double>((*darr));

  for(int i=0;i<r;i++) {
    for(int j=0;j<c;j++) {
      (*darr)[M2A(i,j,r,c)] = (*tdarr)[M2A(i+y,j+x,r,c)];
    }
  }
  delete tdarr;
}

void Pnm::darrclamp(Array<double>* darr,double lo,double hi)
{
  for(int i=0;i<r;i++) {
    for(int j=0;j<c;j++) {
           if((*darr)[M2A(i,j,r,c)] < lo) (*darr)[M2A(i,j,r,c)] = lo;
      else if((*darr)[M2A(i,j,r,c)] > hi) (*darr)[M2A(i,j,r,c)] = hi;
    }
  }
}

void Pnm::darrnorm(Array<double>* darr)
{
	double	min=HUGE,max=-HUGE,denom;

  for(int i=0;i<r;i++) {
    for(int j=0;j<c;j++) {
      if((*darr)[M2A(i,j,r,c)] > max) max = (*darr)[M2A(i,j,r,c)];
      if((*darr)[M2A(i,j,r,c)] < min) min = (*darr)[M2A(i,j,r,c)];
    }
  }
  denom = max - min + 1.;
  for(int i=0;i<r;i++) {
    for(int j=0;j<c;j++) {
      (*darr)[M2A(i,j,r,c)] =
     ((*darr)[M2A(i,j,r,c)] - min)/denom * ((double)maxval);
    }
  }
}

Array<double>* Pnm::dresize(Array<double>* darr,int nr,int nc)
{
	int		lo,hi;
	int		col_expand, row_expand;
	double		w,last,sum;
	double		cols, col_scale, col_sampling_factor;
	double		rows, row_scale, row_sampling_factor;
	Array<double>	*tdarr, *ndarr;

  /* this whole business of _w_, _sum_ and _last_ provide a mechanism
     so that if pixels overlap, the total weights will add up
     to 1.  This is a kind of greedy mechanism that always tries
     to weight a pixel by 1 unless that pixel only contributes
     partially to the new pixel.  In all cases, each pixel
     of the given row must not contribute more than nc/c to the
     expanded pixel(s).
  */

  if(nc/c) {
    col_expand = 1;
    col_scale = (double)((double)nc/(double)c);
    col_sampling_factor = 1.;
    cols = c;
  }
  else {
    col_expand = 0;
    col_scale = (double)((double)c/(double)nc);
    col_sampling_factor = col_scale;
    cols = nc;
  }
  if(nr/r) {
    row_expand = 1;
    row_scale = (double)((double)nr/(double)r);
    row_sampling_factor = 1.;
    rows = r;
  }
  else {
    row_expand = 0;
    row_scale = (double)((double)r/(double)nr);
    row_sampling_factor = row_scale;
    rows = nr;
  }

  tdarr = new Array<double>(r*nc);
  ndarr = new Array<double>(nr*nc);

  /* step 1: shrink/expand columns */
  for(int i=0;i<r;i++) {
    last = 0;
    for(int j=0;j<cols;j++) {
      lo = (int)floor(j*col_scale);
      hi = (int)ceil((j+1)*col_scale);
      w = 1. - last;
      if( (fabs(w-0.) < 0.000001) || (w < 0.) ) w = 1.;
      sum = 0;
      for(int k=lo;k<hi;k++) {
        if(col_expand)
          (*tdarr)[M2A(i,k,r,nc)] +=
            w/col_sampling_factor*(*darr)[M2A(i,j,r,c)];
        else
          (*tdarr)[M2A(i,j,r,nc)] +=
            w/col_sampling_factor*(*darr)[M2A(i,k,r,c)];
        sum += w;
        last = w;
        w = col_scale - sum;
        if(w>1.) w = 1.;
      }
    }
  }

  /* step 2: shrink/expand rows */
  for(int j=0;j<nc;j++) {
    last = 0;
    for(int i=0;i<rows;i++) {
      lo = (int)floor(i*row_scale);
      hi = (int)ceil((i+1)*row_scale);
      w = 1. - last;
      if( (fabs(w-0.) < 0.000001) || (w < 0.) ) w = 1.;
      sum = 0;
      for(int k=lo;k<hi;k++) {
        if(row_expand)
          (*ndarr)[M2A(k,j,nr,nc)] +=
            w/row_sampling_factor*(*tdarr)[M2A(i,j,r,nc)];
        else
          (*ndarr)[M2A(i,j,nr,nc)] +=
            w/row_sampling_factor*(*tdarr)[M2A(k,j,r,nc)];
        sum += w;
        last = w;
        w = row_scale - sum;
        if(w>1.) w = 1.;
      }
    }
  }

  delete tdarr;
  return ndarr;
}
