#ifndef PNM_H
#define PNM_H

#define PBM_WHITE 0
#define PBM_BLACK 1

#define PGM_MAXMAXVAL 255
#define PPM_MAXMAXVAL PGM_MAXMAXVAL

// PBM Magic constants.
#define PBM_MAGIC1 'P'
#define PBM_MAGIC2 '1'
#define RPBM_MAGIC2 '4'
#define PBM_FORMAT (PBM_MAGIC1 * 256 + PBM_MAGIC2)
#define RPBM_FORMAT (PBM_MAGIC1 * 256 + RPBM_MAGIC2)
#define PBM_TYPE PBM_FORMAT

// PGM Magic constants.
#define PGM_MAGIC1 'P'
#define PGM_MAGIC2 '2'
#define RPGM_MAGIC2 '5'
#define PGM_FORMAT (PGM_MAGIC1 * 256 + PGM_MAGIC2)
#define RPGM_FORMAT (PGM_MAGIC1 * 256 + RPGM_MAGIC2)
#define PGM_TYPE PGM_FORMAT

// PPM Magic constants.
#define PPM_MAGIC1 'P'
#define PPM_MAGIC2 '3'
#define RPPM_MAGIC2 '6'
#define PPM_FORMAT (PPM_MAGIC1 * 256 + PPM_MAGIC2)
#define RPPM_FORMAT (PPM_MAGIC1 * 256 + RPPM_MAGIC2)
#define PPM_TYPE PPM_FORMAT

// Macro for turning a format number into a type number.
#define PBM_FORMAT_TYPE(f) \
  ((f) == PBM_FORMAT || (f) == RPBM_FORMAT ? PBM_TYPE : -1)

// Macro for turning a format number into a type number.
#define PPM_FORMAT_TYPE(f) \
  ((f) == PPM_FORMAT || (f) == RPPM_FORMAT ? PPM_TYPE : PGM_FORMAT_TYPE(f))

// Macro for turning a format number into a type number.
#define PGM_FORMAT_TYPE(f) \
  ((f) == PGM_FORMAT || (f) == RPGM_FORMAT ? PGM_TYPE : PBM_FORMAT_TYPE(f))

// Wraparound defines:  note that i,j can be (i+ni), (i-ni), (j+nj), (j-nj)
// which allows the indices i,j to go either left or right in the image row
//
#define	I(i,r)			(((r)+(i))%(r))
#define	J(j,c)			(((c)+(j))%(c))
#define	M2A(i,j,r,c)		(I((i),(r))*(c) + J((j),(c)))

typedef unsigned char gray;
typedef unsigned char bit;
typedef unsigned char byte;

class Pnm {
	public:
	// public member functions
virtual	void	read(const char* fp, const char* fn)	= 0;
virtual	void	write(const char* fp, const char* fn)	= 0;
virtual	void	resize(int nr, int nc)			= 0;
virtual	void	rotate(double angle,int bilinear)	= 0;
virtual void	thresh(double min,double max)		= 0;
	int	rows(void)				{ return(r); }
	int	cols(void)				{ return(c); }
	void		rename(const char* fp,const char* fn)
	  { if(fp) fupdate(fpath,fp);
	    if(fn) fupdate(fname,fn);
	  }
	char*	name(void) { return fname; }
	char*	path(void) { return fpath; }

	// protected: available to this (base) class and subs. derived classes
	protected:
virtual	void	getHeader(istream& s)			= 0;
virtual	void	putHeader(ostream& s)			= 0;
virtual	void    getImage(istream& s)			= 0;
virtual	void    getImage_btot(istream& s)		= 0;
virtual	void    putImage(ostream& s)			= 0;
virtual	void    putImage_btot(ostream& s)		= 0;
	int 	getInt(istream& s);
	char 	getChar(istream& s);
	byte 	getRawByte(istream& s);
	bit 	getBit(istream& s);
	void	putUS(ushort n, ostream& s);
	void	fupdate(char*& deststr,const char* srcstr);

	Array<double>*	dresize(Array<double>*,int nr,int nc);
	void		darrshift(Array<double>*,int x,int y);
	void		darrclamp(Array<double>*,double lo,double hi);
	void		darrnorm(Array<double>*);
	void		darrrot(Array<double>*,double angle,int bilinear);
	void		darrthresh(Array<double>*,double min,double max);

	int		r,c;
	int		format;
	int		maxval;
	char*		fpath;
	char*		fname;

	// private: only available to this class
	private:
};

#endif
