#include	<iostream>
#include	<iomanip>
#include	<fstream>
#include	<sstream>
#include	<string>
#include	<cmath>

using namespace std;

#include	<mlib/types.h>
#include	<mlib/array.h>
#include	"lum.h"
#include	"rgba.h"
#include	"smpte.h"
#include	"pnm.h"
#include	"ppm.h"

/////////////////////////// protected helper functions (for constructors) //////
void Ppm::init(int row, int col, ulong clr)
{
  init(row,col);
  if(!clr) {
    testPattern_btot();
//  testPattern();
    rename("./","smpte.ppm");
  }
}

void Ppm::init(int row, int col)
{
  r		= row;
  c		= col;
  format	= PPM_FORMAT;
  maxval	= 255;
  fpath		= NULL;
  fname		= NULL;
  larr		= new Array<ulong>(r*c);
  drarr		= NULL;
  dgarr		= NULL;
  dbarr		= NULL;
}

void Ppm::init(const Ppm& rhs)
{
  r		= rhs.r;
  c		= rhs.c;
  format	= rhs.format;
  maxval	= rhs.maxval;
  fpath		= rhs.fpath ? strdup(rhs.fpath) : NULL;
  fname		= rhs.fname ? strdup(rhs.fname) : NULL;
  larr		= NULL;
  drarr		= NULL;
  dgarr		= NULL;
  dbarr		= NULL;

  rename(rhs.fpath,rhs.fname);
  larrcpy(rhs.larr);
  darrcpy(rhs.drarr,rhs.dgarr,rhs.dbarr);
}


/////////////////////////// operators: (const Ppm&, const Ppm&) ////////////////
const Ppm& Ppm::operator=(const Ppm& rhs)
{
  return copy(rhs);
}

const Ppm& Ppm::copy(const Ppm& rhs)
{
  if(this != &rhs) {
    if( (r != rhs.r) && (c != rhs.c) ) {
      delete larr;
      delete drarr;  delete dgarr;  delete dbarr;
      r = rhs.r;  c = rhs.c;
    }
    larrcpy(rhs.larr);
    darrcpy(rhs.drarr,rhs.dgarr,rhs.dbarr);
  }
  return *this;
}

const Ppm& Ppm::operator+=(const Ppm& rhs)
{
	int		nr,nc;
	Array<ulong>	*newlarr;

  // THIS IS A FAIRLY QUICK HACK: used in conjunction with bcircf - it's 
  //                              just used to "add" pixels from the given
  //                              image and an overlay that was temporarily
  //                              used to draw filled circle in.
  // assume images are of the same size - normally we'd have to check
  // sizes, create a new array
  if(this != &rhs) {
    if( (r != rhs.r) && (c != rhs.c) ) {
      nr = (r <= rhs.r) ? r : rhs.r;
      nc = (c <= rhs.c) ? c : rhs.c;
      newlarr = new Array<ulong>(nr * nc);
      // copy old image; wraparound indices of new image will take care
      // of size difference, if any (may be somewhat strange behavior though)
      for(int i=0; i < r; i++)
        for(int j=0; j < c; j++)
          (*newlarr)[M2A(i,j,nr,nc)] = (*this)(i,j);
      delete larr;
      larr = newlarr;
      r = nr;  c = nc;
    }
  }

  // if overlay image's pixel is non-black, fill in this image's pixel
  // otherwise do nothing since overlay image is black, or this pixel
  // is already filled
  for(int i=0; i < r; i++)
    for(int j=0; j < c; j++)
      if(rhs(i,j))
        (*this)(i,j) = rhs(i,j); // this size must be >= rhs.size

  return *this;
}


/////////////////////////// friends ////////////////////////////////////////////
ostream& operator<<(ostream& s,Ppm& rhs)
{
  rhs.putHeader(s);
  rhs.putImage_btot(s);
  return s;
}

istream& operator>>(istream& s,Ppm& rhs)
{
  rhs.getHeader(s);
  rhs.getImage_btot(s);
  return s;
}


/////////////////////////// public member functions ////////////////////////////
void Ppm::read(const char* fp,const char* fn)
{
	ostringstream	file;

  rename(fp,fn);

  file << fp << '/' << fn << '\0';

  ifstream ifs(file.str().c_str());
  if(!ifs) {
    cerr << "Can't open input " << file << endl;
    exit(1);
  }

  getHeader(ifs);
  getImage(ifs);

  ifs.close();
}

void Ppm::write(const char* fp,const char* fn)
{
	ostringstream	file;

  rename(fp,fn);

  file << fp << '/' << fn << '\0';

  ofstream ofs(file.str().c_str());
  if(!ofs) {
    cerr << "Can't open output " << file << endl;
    exit(1);
  }

  putHeader(ofs);
  putImage(ofs);

  ofs.close();
}

void Ppm::resize(int nr, int nc)
{
  if( (nr != r) || (nc != c) ) {
    ltod();
    drarr = new Array<double>(*dresize(drarr,nr,nc));
    dgarr = new Array<double>(*dresize(dgarr,nr,nc));
    dbarr = new Array<double>(*dresize(dbarr,nr,nc));
    r = nr; c = nc;
    dtol();
  }
}

void Ppm::rotate(double angle,int bilinear)
{
  ltod();
  darrrot(drarr,angle,bilinear);
  darrrot(dgarr,angle,bilinear);
  darrrot(dbarr,angle,bilinear);
  dtol();
}

void Ppm::thresh(double min,double max)
{
  ltod();
  darrthresh(drarr,min,max);
  darrthresh(dgarr,min,max);
  darrthresh(dbarr,min,max);
  dtol();
}

/////////////////////////// protected member functions /////////////////////////
void Ppm::getHeader(istream& s)
{
	int	ich1, ich2;

  ich1 = s.get();
  if(s.eof()) {
    cerr << "getHeader: EOF / read error reading magic number\n";
    exit(1);
  }
  ich2 = s.get();
  if(s.eof()) {
    cerr << "getHeader: EOF / read error reading magic number\n";
    exit(1);
  }
  format = ich1 * 256 + ich2;

  switch(PPM_FORMAT_TYPE(format)) {
    case PPM_TYPE:
      c = getInt(s);
      r = getInt(s);
      maxval = getInt(s);
      if(maxval > PPM_MAXMAXVAL) {
        cerr << "getHeader: maxval is too large - ";
        cerr << "try reconfiguring with PGM_BIGGRAYS\n";
        exit(1);
      }
      break;
    case PGM_TYPE:
      c = getInt(s);
      r = getInt(s);
      maxval = getInt(s);
      if(maxval > PGM_MAXMAXVAL) {
        cerr << "getHeader: maxval is too large - ";
        cerr << "try reconfiguring with PGM_BIGGRAYS\n";
        exit(1);
      }
      break;
    case PBM_TYPE:
      c = getInt(s);
      r = getInt(s);
      maxval = 1;
      break;
    default:
      cerr << "getHeader: bad magic number - not a pgm or pbm file\n";
      exit(1);
  }
}

void Ppm::putHeader(ostream& s)
{
  s << PPM_MAGIC1 << RPPM_MAGIC2 << endl;
  s << c << " " << r << endl;
  s << (int)maxval << endl;
}

void Ppm::getImage(istream& s)
{
	int		i,j;
	register int	bitshift;
	register gray	g, item;
	register bit	b;
	register uchar	blue,green,red,alpha=0xff;

  if(larr) delete larr;
  larr = new Array<ulong>(r*c);

  for(i=0;i<r;++i) {
    bitshift = -1;
    for(j=0;j<c;j++) {
      switch(format) {
        case PPM_FORMAT:
          red = getInt(s);
          green = getInt(s);
          blue = getInt(s);
          break;
        case RPPM_FORMAT:
          red = getRawByte(s);
          green = getRawByte(s);
          blue = getRawByte(s);
          break;
        case PGM_FORMAT:
          g = getInt(s);
          blue = green = red = (uchar)g;
          break;
        case RPGM_FORMAT:
          g = getRawByte(s);
          blue = green = red = (uchar)g;
          break;
        case PBM_FORMAT:
        case RPBM_FORMAT:
          switch(format) {
            case PBM_FORMAT:
              b = getBit(s);
              break;
            case RPBM_FORMAT:
              if(bitshift == -1) {
                item = getRawByte(s);
                bitshift = 7;
              }
              b = (item >> bitshift) & 1;
              --bitshift;
              break;
            default:
              cerr << "getImage: unknown PBM format\n";
              exit(1);
          }
          g = (b==PBM_WHITE) ? maxval : 0;
          blue = green = red = (uchar)g;
          break;
        default:
          cerr << "getImage: unknown PGM format\n";
          exit(1);
      }
#ifdef ANM_OSX
      // store in RGBA format
      (*this)(i,j) = ((red<<24) | (green<<16) | (blue<<8) | alpha);
#else
      // store in ABGR format (old way; for MESA)
      (*this)(i,j) = ((alpha<<24) | (blue<<16) | (green<<8) | red);
#endif
    }
  }
}

void Ppm::getImage_btot(istream& s)
{
	int		i,j;
	register int	bitshift;
	register gray	g, item;
	register bit	b;
	register uchar	blue,green,red,alpha=0xff;

  if(larr) delete larr;
  larr = new Array<ulong>(r*c);

  for(i=r-1;i>=0;i--) {
    bitshift = -1;
    for(j=0;j<c;j++) {
      switch(format) {
        case PPM_FORMAT:
          red = getInt(s);
          green = getInt(s);
          blue = getInt(s);
          break;
        case RPPM_FORMAT:
          red = getRawByte(s);
          green = getRawByte(s);
          blue = getRawByte(s);
          break;
        case PGM_FORMAT:
          g = getInt(s);
          blue = green = red = (uchar)g;
          break;
        case RPGM_FORMAT:
          g = getRawByte(s);
          blue = green = red = (uchar)g;
          break;
        case PBM_FORMAT:
        case RPBM_FORMAT:
          switch(format) {
            case PBM_FORMAT:
              b = getBit(s);
              break;
            case RPBM_FORMAT:
              if(bitshift == -1) {
                item = getRawByte(s);
                bitshift = 7;
              }
              b = (item >> bitshift) & 1;
              --bitshift;
              break;
            default:
              cerr << "getImage: unknown PBM format\n";
              exit(1);
          }
          g = (b==PBM_WHITE) ? maxval : 0;
          blue = green = red = (uchar)g;
          break;
        default:
          cerr << "getImage: unknown PGM format\n";
          exit(1);
      }
#ifdef ANM_OSX
      // store in RGBA format
      (*this)(i,j) = ((red<<24) | (green<<16) | (blue<<8) | alpha);
#else
      // store in ABGR format (old way; for MESA)
      (*this)(i,j) = ((alpha<<24) | (blue<<16) | (green<<8) | red);
#endif
    }
  }
}

void Ppm::putImage(ostream& s)
{
	int		i,j,charcount=0;
	register gray	g;
	register uchar	blue,green,red;

#ifdef PBMPLUS_RAWBITS
  for(i=0;i<r;i++) {
    for(j=0;j<c;j++) {
      red	= (uchar)LONG_R((*this)(i,j));
      green	= (uchar)LONG_G((*this)(i,j));
      blue	= (uchar)LONG_B((*this)(i,j));
      s.put(red);
      s.put(green);
      s.put(blue);
    }
  }
#else
  for(i=0;i<r;i++) {
    charcount = 0;
    for(j=0;j<c;j++) {
      red	= (uchar)LONG_R((*this)(i,j));
      green	= (uchar)LONG_G((*this)(i,j));
      blue	= (uchar)LONG_B((*this)(i,j));
      putUS((ushort)red,s);
      putUS((ushort)green,s);
      putUS((ushort)blue,s);
      if(charcount >= 65) {
        s.put('\n');
        charcount = 0;
      }
      else if(charcount > 0) {
        s.put(' ');
        s.put(' ');
        charcount += 2;
      }
      putUS((ushort)red,s);
      s.put(' ');
      putUS((ushort)green,s);
      s.put(' ');
      putUS((ushort)blue,s);
      s.put(' ');
      charcount += 11;
    }
    if(charcount > 0)
      s.put('\n');
  }
#endif
}

void Ppm::putImage_btot(ostream& s)
{
	int		i,j,charcount=0;
	register gray	g;
	register uchar	blue,green,red;

#ifdef PBMPLUS_RAWBITS
  for(i=r-1;i>=0;i--) {
    for(j=0;j<c;j++) {
      red	= (uchar)LONG_R((*this)(i,j));
      green	= (uchar)LONG_G((*this)(i,j));
      blue	= (uchar)LONG_B((*this)(i,j));
      s.put(red);
      s.put(green);
      s.put(blue);
    }
  }
#else
  for(i=r-1;i>=0;i--) {
    charcount = 0;
    for(j=0;j<c;j++) {
      red	= (uchar)LONG_R((*this)(i,j));
      green	= (uchar)LONG_G((*this)(i,j));
      blue	= (uchar)LONG_B((*this)(i,j));
      putUS((ushort)red,s);
      putUS((ushort)green,s);
      putUS((ushort)blue,s);
      if(charcount >= 65) {
        s.put('\n');
        charcount = 0;
      }
      else if(charcount > 0) {
        s.put(' ');
        s.put(' ');
        charcount += 2;
      }
      putUS((ushort)red,s);
      s.put(' ');
      putUS((ushort)green,s);
      s.put(' ');
      putUS((ushort)blue,s);
      s.put(' ');
      charcount += 11;
    }
    if(charcount > 0)
      s.put('\n');
  }
#endif
}

void Ppm::testPattern(void)
{
	int	tw=c/7+1,th=2*r/3+1;	// top bars: width, height
	int	bw=c/5+1,bh=1*r/3+1;	// bot bars: width, height

  for(int i=0;i<r;i++) {
    for(int j=0;j<c;j++) {
      if(i <= th) {		// top bars
        switch(j/tw) {
          case SMPTE_TOP_WHITE:   (*this)(i,j) = SMPTE_WHITE_32;	break;
          case SMPTE_TOP_YELLOW:  (*this)(i,j) = SMPTE_YELLOW_32;	break;
          case SMPTE_TOP_CYAN:    (*this)(i,j) = SMPTE_CYAN_32;		break;
          case SMPTE_TOP_GREEN:   (*this)(i,j) = SMPTE_GREEN_32;	break;
          case SMPTE_TOP_MAGENTA: (*this)(i,j) = SMPTE_MAGENTA_32;	break;
          case SMPTE_TOP_RED:     (*this)(i,j) = SMPTE_RED_32;		break;
          case SMPTE_TOP_BLUE:    (*this)(i,j) = SMPTE_BLUE_32;		break;
        }
      }
      else {			// bot bars
        switch(j/bw) {
          case SMPTE_BOT_BLUE1:  (*this)(i,j) = SMPTE_BLUE_32;		break;
          case SMPTE_BOT_WHITE:  (*this)(i,j) = SMPTE_100WHITE_32;	break;
          case SMPTE_BOT_BLUE2:  (*this)(i,j) = SMPTE_BLUE_32;		break;
          case SMPTE_BOT_BLACK1: (*this)(i,j) = SMPTE_BLACK_32;		break;
          case SMPTE_BOT_BLACK2: (*this)(i,j) = SMPTE_BLACK_32;		break;
        }
      }
    }
  }
}

void Ppm::testPattern_btot(void)
{
	int	tw=c/7+1,th=2*r/3+1;	// top bars: width, height
	int	bw=c/5+1,bh=1*r/3+1;	// bot bars: width, height

  for(int i=r-1;i>=0;i--) {
    for(int j=0;j<c;j++) {
      if(i >= bh) {		// top bars
        switch(j/tw) {
          case SMPTE_TOP_WHITE:   (*this)(i,j) = SMPTE_WHITE_32;	break;
          case SMPTE_TOP_YELLOW:  (*this)(i,j) = SMPTE_YELLOW_32;	break;
          case SMPTE_TOP_CYAN:    (*this)(i,j) = SMPTE_CYAN_32;		break;
          case SMPTE_TOP_GREEN:   (*this)(i,j) = SMPTE_GREEN_32;	break;
          case SMPTE_TOP_MAGENTA: (*this)(i,j) = SMPTE_MAGENTA_32;	break;
          case SMPTE_TOP_RED:     (*this)(i,j) = SMPTE_RED_32;		break;
          case SMPTE_TOP_BLUE:    (*this)(i,j) = SMPTE_BLUE_32;		break;
        }
      }
      else {			// bot bars
        switch(j/bw) {
          case SMPTE_BOT_BLUE1:  (*this)(i,j) = SMPTE_BLUE_32;		break;
          case SMPTE_BOT_WHITE:  (*this)(i,j) = SMPTE_100WHITE_32;	break;
          case SMPTE_BOT_BLUE2:  (*this)(i,j) = SMPTE_BLUE_32;		break;
          case SMPTE_BOT_BLACK1: (*this)(i,j) = SMPTE_BLACK_32;		break;
          case SMPTE_BOT_BLACK2: (*this)(i,j) = SMPTE_BLACK_32;		break;
        }
      }
    }
  }
}

/////////////////////////// private member functions ///////////////////////////
void Ppm::ltod(void)
{
	double		dr,dg,db;
	register uchar	red,green,blue;

  if(!larr) return;

  if(drarr) delete drarr;  if(dgarr) delete dgarr;  if(dbarr) delete dbarr;
  drarr = new Array<double>(r*c);
  dgarr = new Array<double>(r*c);
  dbarr = new Array<double>(r*c);

  for(int i=0;i<r;i++) {
    for(int j=0;j<c;j++) {
      red       = (uchar)LONG_R((*this)(i,j));
      green     = (uchar)LONG_G((*this)(i,j));
      blue      = (uchar)LONG_B((*this)(i,j));
      dr = (double)red; dg = (double)green; db = (double)blue;
      (*drarr)[M2A(i,j,r,c)] = dr;
      (*dgarr)[M2A(i,j,r,c)] = dg;
      (*dbarr)[M2A(i,j,r,c)] = db;
    }
  }
}

void Ppm::dtol(void)
{
	register uchar	red,green,blue;

  if(!drarr || !dgarr || !dbarr) return;

  if(larr) delete larr;
  larr = new Array<ulong>(r*c);

  for(int i=0;i<r;i++) {
    for(int j=0;j<c;j++) {
      red	= static_cast<uchar>((*drarr)[M2A(i,j,r,c)]);
      green	= static_cast<uchar>((*dgarr)[M2A(i,j,r,c)]);
      blue	= static_cast<uchar>((*dbarr)[M2A(i,j,r,c)]);
      (*this)(i,j) = ((red<<24) | (green<<16) | (blue<<8));
    }
  }
}

/////////////////////////// public member functions : circle drawing ///////////
void Ppm::fill8(int i,int j)
{
	int	ni,nj;

  if((*this)(i,j) != 0x77777777) {
    setPixel(i,j);
    for(ni=-1;ni<=1;ni++) {
      for(nj=-1;nj<=1;nj++) {
        fill8(i+ni,j+nj);
      }
    }
  }
}

void Ppm::fill4(int i,int j)
{
	int	ni,nj;

  if((*this)(i,j) != 0x77777777) {
    setPixel(i,j);
    fill4(i,j-1);
    fill4(i,j+1);
    fill4(i-1,j);
    fill4(i+1,j);
  }
}

void Ppm::bcircf(int cx,int cy,int radius,int thk)
{
  bcirc(cx,cy,radius,thk);
  // don't use fill8 on Breshenham's circle because it's only 4-connected;
  // fill8 will overflow and cover the entire image.
  fill4(cx,cy);
}

void Ppm::bcirc(int cx,int cy,int radius,int thk)
{
        int    x,y,p;
 
  // Draws circle of given radius and thickness at center       \ 7 | 0 /   
  // (cx,cy).  Bcirc() uses Bresenham's aglorithm for drawing    \  |  /    
  // the primary arc in octant 0 (see diagram on right).          \ | /     
  //                                                            6  \|/  1   
  // The primary coordinates are them mapped into their        -----|------ 
  // corresponding positions.                                   5  /|\  2   
  //                                                              / | \     
  // For circles of thickness thk < 5 pixels are added to        /  |  \    
  // the primary pixel on alternating y+1, y-1 sides.           / 4 | 3 \   
  //
  p = 3 - (2*radius);
  x = 0;  y = radius;
  while(x < y) {
    drawCircPixel(x,y,cx,cy,thk);
    if(p<0) 
      p = p + (4 * x) + 6;
    else {
      p = p + (4 * (x - y)) + 10;
      y = y - 1;
    }
    x = x + 1;
  }
  if(x == y) drawCircPixel(x,y,cx,cy,thk);
}
 
void Ppm::drawCircPixel(int x,int y,int cx,int cy,int thk)
{
	int	i;

  // Draws circles of given thickness thk < 5.
  for(i=1;i<=thk;i++) {
    if(i%2 != 0)	y = y + (i/2);
    else		y = y - (i/2);
    setCircPixel(x,y,cx,cy);
  }
} 

void Ppm::setCircPixel(int x,int y,int cx,int cy)
{
  // Map pixels to octants 0 - 7.
  setPixel(cx + x,cy + y);
  setPixel(cx - x,cy + y);
  setPixel(cx + x,cy - y);
  setPixel(cx - x,cy - y);
  setPixel(cx + y,cy + x);
  setPixel(cx - y,cy + x);
  setPixel(cx + y,cy - x);
  setPixel(cx - y,cy - x);
}

void Ppm::setPixel(int i,int j)
{
  // set RGBA pixel inverting bgr fields with OR, preserving alpha channel
  (*this)(i,j) = 0x77777777;
}
