#ifndef PPM_H
#define PPM_H

class Ppm : public Pnm {
	public:
	// constructors (overloaded)
	Ppm(int row=256,int col=256,ulong clr=0)	{init(row,col,clr);}
	Ppm(int row=256,int col=256)			{init(row,col);}
	Ppm(const Ppm& rhs) : Pnm(rhs)			{init(rhs);}
	Ppm(Ppm* rhs)					{init(*rhs);}

	// destructors
virtual	~Ppm()
	{
	  if(larr)	delete larr;
	  if(drarr)	delete drarr;
	  if(dgarr)	delete dgarr;
	  if(dbarr)	delete dbarr;
	  if(fpath)	delete [] fpath;
	  if(fname)	delete [] fname;
	}

	// channel accessors
	ulong& operator()(int i, int j)
	  { return((*larr)[M2A(i,j,r,c)]); }
	ulong& operator()(int i, int j) const
	  { return((*larr)[M2A(i,j,r,c)]); }

	// operators: (const Ppm&, const Ppm&)
	const Ppm& operator=(const Ppm& rhs);
	const Ppm& operator+=(const Ppm& rhs);
	const Ppm& operator+=(Ppm* rhs)		{ return((*this) += (*rhs)); }

	// operators: (const Ppm&, Ppm*)
	const Ppm& operator=(Ppm* rhs)
	  { return((*this)  = (*rhs)); }

	const Ppm& copy(const Ppm& rhs);

	// friends
friend	ostream&	operator<<(ostream& s, Ppm& rhs);
friend	ostream&	operator<<(ostream& s, Ppm *rhs)
	  { return(s << (*rhs)); }
friend	istream&	operator>>(istream& s, Ppm& rhs);
friend	istream&	operator>>(istream& s, Ppm* rhs)
	  { return(s >> (*rhs)); }

	// public member functions
virtual	void	read(const char* fp, const char* fn);
virtual	void	write(const char* fp, const char* fn);
virtual	void	resize(int nr, int nc);
virtual	void	rotate(double angle,int bilinear);
virtual void	thresh(double min,double max);

	void	black()
	  { if(larr) larr->black(); }

	void	fill8(int i,int j);
	void	fill4(int i,int j);
	void	bcircf(int cx,int cy,int radius,int thk);
	void	bcirc(int cx,int cy,int radius,int thk);
	void	drawCircPixel(int x,int y,int cx,int cy,int thk);
	void	setCircPixel(int x,int y,int cx,int cy);
	void	setPixel(int i,int j);

	ulong*	data()
	  { return(larr->data()); }

	// protected: available to this (base) class and subs. derived classes
	protected:
	void	init(int row,int col);
	void	init(int row,int col,ulong clr);
	void	init(const Ppm& rhs);
virtual	void	getHeader(istream& s);
virtual	void	putHeader(ostream& s);
virtual	void	getImage(istream& s);
virtual	void	getImage_btot(istream& s);
virtual	void	putImage(ostream& s);
virtual	void	putImage_btot(ostream& s);
	void	testPattern(void);
	void	testPattern_btot(void);

	void	larrcpy(Array<ulong>* slr)
	  { if(slr) larr = new Array<ulong>((*slr)); }

	void	dtol(void);
	void	ltod(void);

	void	darrcpy(Array<double>* sdr,Array<double>* sdg,Array<double>* sdb)
	  {
  	    if(sdr) drarr = new Array<double>((*sdr));
  	    if(sdg) dgarr = new Array<double>((*sdg));
  	    if(sdb) dbarr = new Array<double>((*sdb));
	  }

	Array<ulong>*	larr;
	Array<double>*	drarr;
	Array<double>*	dgarr;
	Array<double>*	dbarr;

	// private: only available to this class
	private:
};

#endif
