#include	<iostream>
#include        <iomanip>
#include        <string>
#include	<math.h>

using namespace std;

#include	"types.h"
#include	"array.h"

/////////////////////////// protected helper functions (for constructors) //////
template <typename Type>
void Array<Type>::init(int sz, Type *array)
{
  // init routine:
  //   may be called with array == NULL,
  //   but sz will always have valid value (default = 5)

  size = sz;			// set private data member 'size'
  arr = new Type [size];	// allocate storage for 'arr'

  // if array == NULL, just set array to zero,
  // otherwise copy data from array
  if(array)	copy(array);
  else		memset(arr,0,size*sizeof(Type));
}

template <typename Type>
inline void Array<Type>::copy(Type *array)
{
  // arr must have been created,
  // size must have been set,
  // so just copy all elements from array to arr
  for(int i=0; i < size; ++i)
    arr[i] = array[i];
}

/////////////////////////// operators: (const Array&, const Array&) ////////////
template <typename Type>
const Array<Type>& Array<Type>::operator=(const Array& rhs)
{
  if(this != &rhs) {                            // compare _pointers_ !
    if(size != rhs.size) {
      delete [] arr;                            // arr already exists, nuke it!
      size = rhs.size;
      arr = new Type [size];
    }
    copy(rhs.arr);
  }
  return *this;
}

template <typename Type>
Array<Type> Array<Type>::operator+=(const Array& rhs)
{
	Type	*newarr;

  if(this->size <= rhs.size) {
    newarr = new Type [rhs.size];
    memset(newarr,0,rhs.size*sizeof(Type));
    for(int i=0; i < this->size; i++)
      newarr[i] = this->arr[i];
    delete [] this->arr;
    this->size = rhs.size;
    this->arr = newarr;
  }

  for(int i=0; i < rhs.size; i++)
    this->arr[i] += rhs.arr[i];		// this->size must be >= rhs.size

  return *this;
}

template <typename Type>
Array<Type> Array<Type>::operator+(const Array& rhs)
{
  // 1. create an Array object with large enough array to hold larger of
  //    of the two operands (we'll effectively pad the smaller with 0's).
  // 2. copy the larger arrays's contents

  if(this->size <= rhs.size) {
        Array<Type>    result(rhs);

    return(result += this);
  }
  else {
        Array<Type>    result(this);

    return(result += rhs);
  }
}

/////////////////////////// friends ////////////////////////////////////////////
template <typename Type>
ostream& operator<<(ostream& s,const Array<Type>& rhs)
{
  s.setf(ios::fixed,ios::floatfield);
  s.precision(2);
  for(int i=0; i<rhs.size; i++) {
    s.width(6);
    i == rhs.size-1 ? s << rhs[i] : s << rhs[i] << ", ";
  }
  return s;
}

/////////////////////////// public member functions ////////////////////////////
template <typename Type>
Type Array<Type>::min() const
{
	Type	min = arr[0];

  // return minimum value
  for(int i=0; i < size; i++)
    if(arr[i] < min) min = arr[i];

  return min;
}

template <typename Type>
Type Array<Type>::max() const
{
	Type max = arr[0];

  // return maximum value
  for(int i=0; i < size; i++)
    if(arr[i] > max) max = arr[i];

  return max;
}

template <typename Type>
int Array<Type>::linsearch(Type val) const
{
  // if given val is found within the array using linear search,
  //   return the index of its first occurrence
  //   otherwise, return -1

  for(int i=0; i < size; i++)
    if(arr[i] == val) return i;

  return -1;
}

template <typename Type>
int Array<Type>::binsearch(Type val) const
{
  // if given val is found within the array using binary search,
  //   return the index of its first occurrence
  //   otherwise, return -1
  // caveat: array must already be sorted

	int	lo,mi,hi;

  lo = 0;  hi = size-1;
  while(lo <= hi) {
    mi = (lo + hi)/2;
    if(arr[mi] < val)
      lo = mi + 1;
    else if(arr[mi] > val)
      hi = mi - 1;
    else
      return mi;
  }

  return -1;
}

template <typename Type>
void Array<Type>::inssort()
{
	int	i,j;
	Type	tmp;

  // sort the array using insertion sort
  for(i = 1; i < size; i++) {
    tmp = arr[i];
    for(j = i; j > 0 && tmp < arr[j-1]; j--)
      arr[j] = arr[j-1];
    arr[j] = tmp;
  }
}
/////////////////////////// private member functions ///////////////////////////
