#ifndef ARRAY_H
#define ARRAY_H

// Wraparound defines:  note that i,j can be (i+ni), (i-ni), (j+nj), (j-nj)
// which allows the indices i,j to go either left or right in the image
// row
#define	I(i,r)			(((r)+(i))%(r))
#define	J(j,c)			(((c)+(j))%(c))
#define	M2A(i,j,r,c)		(I((i),(r))*(c) + J((j),(c)))

template <typename Type>
class Array {
	public:
	// constructors (overloaded)
	Array(int sz = 5)			{ init(sz,NULL); }
	Array(Type *array, int sz)		{ init(sz,array); }
	Array(const Array& rhs)			{ init(rhs.size,rhs.arr); }
	Array(Array *rhs)			{ init(rhs->size,rhs->arr); }

	// destructors
	~Array() { delete [] arr; }

	// operators
	const Array<Type>& operator=(const Array& rhs);
	Array<Type> operator+=(const Array& rhs);
	Array<Type> operator+(const Array& rhs);
	Type& operator[](int index)		{ return(arr[index]); }
	Type& operator[](int index) const	{ return(arr[index]); }

	// friends
	friend ostream& operator<<(ostream& s, const Array& rhs);
	friend ostream& operator<<(ostream& s, Array *rhs)
	  { return(s << (*rhs)); }

	// member functions
	Type		min() const;
	Type		max() const;
	Type*		data() { return arr; }
	int		linsearch(Type val) const;
	int		binsearch(Type val) const;
	void		inssort();
	void		black()
	  { memset(arr,0,size*sizeof(Type)); }

	// private: only available to this class
	private:
	int		size;
	Type		*arr;

	// protected: available to this (base) class and subs. derived classes
	protected:
	void		init(int sz,Type *array);
	inline void	copy(Type *array);
};

#endif
