#include        <iostream>
#include        <iomanip>
#include	<math.h>

using namespace std;

#include	"types.h"
#include	"array.h"
#include	"matrix.h"
#include	"vector.h"
#include	"quaternion.h"

// prototypes
int	main(int argc, char **argv);
void	test_array();
void	test_quat_mem();
void	test_quaternion();
void	test_vector();
void	test_matrix();

int main(int argc, char **argv)
{
//test_quat_mem();
  for(int i=0; i<1000000; i++) {
  test_array();
  test_quaternion();
  test_matrix();
  test_vector();
  }
}

void test_array()
{
	int		n=10;
	Array<int>	iarr1(n);
	Array<int>	iarr2(n);

  for(int i=0;i<n;i++) iarr2[i] = i;

  cout << iarr1 << endl;
  cout << iarr2 << endl;

  cout << iarr1 + iarr2 << endl;

}

void test_quat_mem()
{
	float		rangle = 45.0*M_PI/180.0;	// theta (in radians)
	Vector<float>	x(1.0,0.0,0.0);			// rotation axis
	Vector<float>	y(0.0,1.0,0.0);			// rotation axis
	Vector<float>	z(0.0,0.0,1.0);			// rotation axis
	Vector<float>	P(10.0,0.0,0.0);		// position (x,y,z)
	Vector<float>	Q(1.0,0.0,0.0);			// position (x,y,z)
	Quat<float>	q((float)(cos(rangle/2.0)),(float)(sin(rangle/2.0))*z);
	Quat<float>	p(0.0,P);

  for(int i=0; i<1000000; i++) {
    p.s() = 0.0;  p.v() = P;
    p * q;
    p = !q * p * q;
    Q = p.v();
    cout << " Q = " << Q << endl;
    Q = Q + P;
  }
}

void test_quaternion()
{

  // typical usage would be:
  //   0. store q with some graphical object
  //   1. in (interactive) loop:
  //     1.1. figure out incremental rotation, q'
  //       1.1.1. (use q' = new Quat<Type>(x1,y1,x2,y2))
  //     1.2. "add to" object's quaternion, q = q' * q (order is important!)
  //     1.3. normalize q, q.norm()
  //   2. in display routine:
  //     2.1. get transformation matrix, M = q.rotMat() (in row-major order!)
  //     2.3. feed M to OpenGL's object matrix stack (may need to transpose)
  //

	float		rangle = 45.0*M_PI/180.0;	// theta (in radians)
	Vector<float>	vzero(0.0,0.0,0.0);			// rotation axis
	Vector<float>	x(1.0,0.0,0.0);			// rotation axis
	Vector<float>	y(0.0,1.0,0.0);			// rotation axis
	Vector<float>	z(0.0,0.0,1.0);			// rotation axis

	Quat<float>	q_0(1.0,vzero);
	Quat<float>	q_1(1.0,x);
	Quat<float>	q_2(1.0,y);
	Quat<float>	q_3(1.0,z);
	Quat<float>	q_4(3.0 * q_1);

	Quat<float>	qx((float)(cos(rangle/2.0)),(float)(sin(rangle/2.0))*x);
	Quat<float>	qy((float)(cos(rangle/2.0)),(float)(sin(rangle/2.0))*y);
	Quat<float>	qz((float)(cos(rangle/2.0)),(float)(sin(rangle/2.0))*z);

	Vector<float>	P(10.0,0.0,0.0);		// position (x,y,z)
	Vector<float>	Ph(10.0,0.0,0.0,1.0);		// position (x,y,z,w)
	Quat<float>	q((float)(cos(rangle/2.0)),(float)(sin(rangle/2.0))*z);
	Quat<float>	p(0.0,P);
//	Matrix<float>	M( (!q).R() * q.L() );		// rotation matrix
	Matrix<float>	M(q.quattomat());
	Quat<float>	qm(M);

	Quat<float>	sq(q_1.slerp(q_2,.5));

  cout << " q_0 = " <<  q_0 << endl;
  cout << " q_0.quattomat() = \n" << q_0.quattomat() << endl;
  cout << " q_1 = " <<  q_1 << endl;
  cout << " q_2 = " <<  q_2 << endl;
  cout << " q_3 = " <<  q_3 << endl;
  cout << " q_1.slerp(q_2,.5) = " <<  sq << endl;
  cout << "-q_1 = " << -q_1 << endl;
  cout << " q_1 * 3.0 = " << q_1 * 3.0 << endl;
  cout << " 3.0 * q_1 = " << 3.0 * q_1 << endl;
  cout << " q_1 * q_2 = " << q_1 * q_2 << endl;
  cout << " q_1 * R(q_2) = " << q_1 * q_2.R() << endl;
  cout << " q_2 * L(q_1) = " << q_2 * q_1.L() << endl;
  cout << " q_4 = " <<  q_4 << endl;
  cout << " |q4| = " << q_4.length() << endl;
  cout << " q_4/|q_4| = " << q_4.norm() << endl;
  cout << " !q_4 = " << !q_4 << endl;
  cout << " q_4 * !q_4 = " << q_4 * !q_4 << endl;
  cout << " q = " <<  q << endl;
  cout << " q * p * !q = " << q * p * !q << endl;
  cout << " (!q).R() * q.L() = \n" << (!q).R() * q.L() << endl;
  cout << " q.L() * (!q).R() = \n" << q.L() * (!q).R() << endl;
  cout << " M = \n" << M << endl;
  cout << " Ph * M = " << Ph * M << endl;
  cout << " qx.quattomat() = \n" << qx.quattomat() << endl;
  cout << " qy.quattomat() = \n" << qy.quattomat() << endl;
  cout << " qz.quattomat() = \n" << qz.quattomat() << endl;
  cout << " qm = " <<  qm << endl;
}

void test_vector()
{
	Vector<float>	vin(0.,0.,0.);
	Vector<float>	v_1(10., 0., 0.);
	Vector<float>	v_2( 0.,10., 0.);
	Vector<float>	v_3( 0., 0., 0.);
	Vector<float>	v_4( 1., 2., 3., 4.);
	Vector<float>	v_1n, v_2n;
	Matrix<float>	m_7(3,3);

  cout << "v_1 = " << v_1 << endl << "v_2 = " << v_2 << endl;
  cout << "v_1[0] = " << v_1[0] << endl;
  cout << "v_1[1] = " << v_1[1] << endl;
  cout << "v_1[2] = " << v_1[2] << endl;
  cout << "v_1 . v_2 = " << v_1.dot(v_2) << endl;
  v_1n = v_1.norm();  v_2n = v_2.norm();
  cout << "v_1/|v_1| = " << v_1n << endl;
  cout << "v_2/|v_2| = " << v_2n << endl;
  cout << "v_1/|v_1| = " << v_1.norm() << endl;
  cout << "v_2/|v_2| = " << v_2.norm() << endl;
  cout << "v_1 = " << v_1 << endl << "v_2 = " << v_2 << endl;
  cout << "v_1 x v_2 = " << v_1.cross(v_2) << endl;
  cout << "v_1n = " << v_1n << endl;
  cout << "v_2n = " << v_2n << endl;
  cout << "v_1/|v_1| x v_2/|v_2| = " << v_1n.cross(v_2n) << endl;
  cout << "v_1 . v_2 = " << v_1.dot(v_2) << endl;
  cout << "v_1 + v_2 = " << v_1 + v_2 << endl;
  cout << "-v_1 = " << -v_1 << endl;
  v_3[0] = -1.; v_3[1] = -1.; v_3[2] = -1.;
  cout << "v_3 = " << v_3 << endl;
  cout << "v_4 = " << v_4 << endl;
  cout << "v_3 + v_4 = " << v_3 + v_4 << endl;
  cout << "v_4 + v_3 = " << v_4 + v_3 << endl;
  m_7(0,0) = 1; m_7(0,1) = 2; m_7(0,2) = 3;
  m_7(1,0) = 2; m_7(1,1) = 3; m_7(1,2) = 1;
  m_7(2,0) = 3; m_7(2,1) = 2; m_7(2,2) = 1;
  cout << "m_7 =" << endl << m_7;
  cout << "v_3 * m_7 = " << v_3 * m_7 << endl;
  v_3 *= m_7;
  cout << "v_3 = " << v_3 << endl;

  cout << endl;

//cout << "Enter in vector (format: x [y [z]])" << endl;
//cin >> vin;
//cout << vin << endl;
}

void test_matrix()
{
	Matrix<float>	m_1(2,2),m_2(2,2),m_3(2,2);
	Matrix<float>	m_4(1,3),m_5(3,3),m_6(1,3);
	Matrix<float>	m_7(3,3);
	Matrix<float>	*m_8 = new Matrix<float>(3,3);

  m_1.identity();  m_2.identity();

  // should be 2x2 identity matrices
  cout << "m_1, m_2 = " << endl;
  cout << m_1 << endl << m_2 << endl;

  // should be 2x2 0 matrix
  cout << "m_1 - m_2 = " << endl;
  cout << m_1 - m_2 << endl;

  // should be 2x2 negative identity matrix
  m_3 = -m_1;
  cout << "m_3 = -m_1" << endl;
  cout << m_3 << endl;

  m_3(0,0) = 1; m_3(0,1) = 2;
  m_3(1,0) = 3; m_3(1,1) = 4;

  m_4(0,0) = 2; m_4(0,1) = 2; m_4(0,2) = 2;

  m_5(0,0) = 1; m_5(0,1) = 2; m_5(0,2) = 3;
  m_5(1,0) = 2; m_5(1,1) = 3; m_5(1,2) = 1;
  m_5(2,0) = 3; m_5(2,1) = 2; m_5(2,2) = 1;

  // should be 1x3 matrix
  m_6 = m_4 * m_5;
  cout << "m_6 = m_4 * m_5" << endl;
  cout << m_6 << endl;

  // should be 1x3 and 3x3 matrices
  cout << "m_4, m_5 = " << endl;
  cout << m_4 << endl << m_5 << endl;

  // should be new 1x3 matrix
  m_4 *= m_5;
  cout << "m_4 *= m_5" << endl;
  cout << m_4 << endl;

  // should be 3x3 inverse matrix
  m_7 = !m_5;
  cout << "m_7 = !m_5" << endl;
  cout << m_7 << endl;

  // should be 3x3 identity matrix
  cout << "m_5 * m_7" << endl;
  cout << m_5 * m_7 << endl;

  // should be 3x3 matrix
  cout << "m_5 =" << endl;
  cout << m_5 << endl;

  // should be 3x3 matrix
  m_5 = m_5 - 2.0;
  cout << "m_5 = m_5 - 2.0" << endl;
  cout << m_5 << endl;

  // should be 3x3 matrix
  cout << "2.0 + m_5" << endl;
  cout << 2.0 + m_5 << endl;

  // should be 3x3 matrix
  m_7 = 2.0 + m_5;
  cout << "m_7 = 2.0 + m_5" << endl;
  cout << m_7 << endl;

  // should be 3x3 matrix
  cout << "2.0 - m_5" << endl;
  cout << 2.0 - m_5 << endl;

  // should be 3x3 matrix
  m_7 = 2.0 - m_5;
  cout << "m_7 = 2.0 - m_5" << endl;
  cout << m_7 << endl;

  // should be 3x3 matrix
  cout << "2.0 * m_5" << endl;
  cout << 2.0 * m_5 << endl;

  // should be 3x3 matrix
  m_7 = 2.0 * m_5;
  cout << "m_7 = 2.0 * m_5" << endl;
  cout << m_7 << endl;

  // should be 3x3 matrix
  cout << "2.0 / m_5" << endl;
  cout << 2.0 / m_5 << endl;

  // should be 3x3 matrix
  m_7 = 2.0 / m_5;
  cout << "m_7 = 2.0 / m_5" << endl;
  cout << m_7 << endl;

  // should be 3x3 0 matrix
  cout << "m_8 =" << endl;
  cout << m_8 << endl;

  delete m_8;

  cout << endl;
}
