#include        <iostream>
#include        <iomanip>
#include	<math.h>

using namespace std;

#include	"matrix.h"
#include	"vector.h"
#include	"quaternion.h"

/////////////////////////// operators: unary ///////////////////////////////////
template <typename Type>
Quat<Type> Quat<Type>::operator-(void)
{
        Quat<Type>    result(0.0);

  // create 0 quaternion, subtract this quaternion from 0 to get -ve quaternion

  return(result -= this);
}

template <typename Type>
Quat<Type> Quat<Type>::operator!(void)
{
	Type		sval = static_cast<Type>(1.0)/dot(this);
        Quat<Type>	result(0.0);

  // let q = (*this) = (s,v), then quaternion inverse is
  //   q^{-1} = \frac{1}{|q|^2} * (s,-v)
  //   where |q|^2 = s^2 + (v . v), e.g., the quaternion's length squared
  //                                      (or its dot product, e.g., dot(this))
  //
  result.scalar = sval *  scalar;
  result.vector = sval * -vector;

  return(result);
}

/////////////////////////// operators: (const Quat&, const Quat&) //////////////
template <typename Type>
int Quat<Type>::operator==(const Quat<Type>& rhs)
{
	float	precision=0.01;

  if(this == &rhs)
    return 1;
  else {
    if(fabs(scalar - rhs.scalar) > precision)
      return 0;
  }
  return(vector == rhs.vector);	// if we get here, lhs.scalar == rhs.scalar
}

template <typename Type>
int Quat<Type>::operator<(const Quat<Type>& rhs)
{
  if(this == &rhs)
    return 0;
  else {
    if(scalar >= rhs.scalar)
      return 0;
  }

  return (vector < rhs.vector); // if we get here, lhs.scalar < rhs.scalar
}

template <typename Type>
int Quat<Type>::operator>(const Quat<Type>& rhs)
{
  if(this == &rhs)
    return 0;
  else {
    if(scalar <= rhs.scalar)
      return 0;
  }

  return (vector > rhs.vector); // if we get here, lhs.scalar > rhs.scalar
}

template <typename Type>
Quat<Type> Quat<Type>::operator=(const Quat<Type>& rhs)
{
  if(this != &rhs) {
    scalar = rhs.scalar;
    vector = rhs.vector;
  }

  return *this;
}

template <typename Type>
Quat<Type> Quat<Type>::operator*=(const Quat<Type>& rhs)
{
  return((*this) = (*this) * rhs);
}

template <typename Type>
Quat<Type> Quat<Type>::operator*(const Quat<Type>& rhs)
{
	Quat<Type>	result(0.0);
  	Vector<Type>	v1(0.0,0.0,0.0), v2(0.0,0.0,0.0), vc(0.0,0.0,0.0);

  // let q_1 = (*this) = (s_1,v_1), q_2 = rhs = (s_2,v_2), then
  //   q_1 q_2 = (s_1 s_2 - v_1.dot(v_2), s_1 v_2 + s_2 v_1 + v_1.cross(v_2))
  //             (    <scalar>          ,         <vector>                  )
  //
  // Note: should also be able to obtain this using the L() and R() routines,
  //       e.g., (*result) = rhs * L(); or (*result) = (*this) * rhs.R();
  //             ONLY IFF class Quat were stored as a 1x4 vector, [x y z s]
  //

  v1 = vector;
  v2 = rhs.vector;

  result.scalar = scalar * rhs.scalar - vector.dot(rhs.vector);

  v1 *= rhs.scalar;
  v2 *=     scalar;
  vc = vector.cross(rhs.vector);

  result.vector = v1 + v2;
  result.vector += vc;

  return result;
}

template <typename Type>
Quat<Type> Quat<Type>::operator*(const Matrix<Type>& rhs)
{
	register Type	s = scalar;
	register Type	x = vector[0], y = vector[1], z = vector[2];
	Vector<Type>	v(x,y,z,s);
	Vector<Type>	vm(v * rhs);	// NOTE: v.c must equal rhs.r
					//       e.g., rhs must be 4x4
	Quat<Type>	q(0.0);

  q.vector[0]	= vm[0];	// x
  q.vector[1]	= vm[1];	// y
  q.vector[2]	= vm[2];	// z
  q.scalar	= vm[3];	// s

  return q;
}

template <typename Type>
Quat<Type> Quat<Type>::operator/=(Quat<Type>& rhs)
{
  return((*this) = (*this) * !rhs);
}

template <typename Type>
Quat<Type> Quat<Type>::operator/(Quat<Type>& rhs)
{
	Quat<Type>	result(0.0);

  return(result = (*this) * !rhs);
}

template <typename Type>
Quat<Type> Quat<Type>::operator+=(const Quat<Type>& rhs)
{

  scalar += rhs.scalar;
  vector += rhs.vector;

  return *this;
}

template <typename Type>
Quat<Type> Quat<Type>::operator+(const Quat<Type>& rhs)
{
	Quat<Type>	result(this);

  return(result += rhs);
}

template <typename Type>
Quat<Type> Quat<Type>::operator-=(const Quat<Type>& rhs)
{
  scalar -= rhs.scalar;
  vector -= rhs.vector;
  
  return *this;
}

template <typename Type>
Quat<Type> Quat<Type>::operator-(const Quat<Type>& rhs)
{
	Quat<Type>	result(this);

  return(result -= rhs);
}

/////////////////////////// operators: (const Quat&, scalar) ///////////////////
template <typename Type>
Quat<Type> Quat<Type>::operator*=(const Type& rhs)
{
  scalar *= rhs;
  vector *= rhs;

  return *this;
}

template <typename Type>
Quat<Type> Quat<Type>::operator*(const Type& rhs)
{
	Quat<Type>	result(this);

  return(result *= rhs);
}

template <typename Type>
Quat<Type> Quat<Type>::operator/=(const Type& rhs)
{
  scalar /= rhs;
  vector /= rhs;

  return *this;
}

template <typename Type>
Quat<Type> Quat<Type>::operator/(const Type& rhs)
{
	Quat<Type>	result(this);

  return(result /= rhs);
}

/////////////////////////// friend operators: (const scalar&, cosnt Quat&) /////
template <typename Type>
Quat<Type> operator*(Type lhs,Quat<Type>& rhs)
{
        Quat<Type>    result(rhs);

  return(result *= lhs);
}

template <typename Type>
Quat<Type> operator/(Type lhs,Quat<Type>& rhs)
{
        Quat<Type>    result(!rhs);

  return(result *= lhs);
}

/////////////////////////// friends ////////////////////////////////////////////
template <typename Type>
ostream& operator<<(ostream& s,const Quat<Type>& rhs)
{
  s.setf(ios::fixed,ios::floatfield);
  s.precision(2);
  s.width(6);
  s << "(" << rhs.scalar << ", " << rhs.vector << ")";

  return s;
}

/////////////////////////// public member functions ////////////////////////////
template <typename Type>
Type Quat<Type>::dot(Quat& rhs)
{
  return( scalar * rhs.scalar + vector.dot(rhs.vector) );
}

template <typename Type>
Matrix<Type> Quat<Type>::L(void)
{
	Matrix<Type>	L_q(4,4);
	register Type	s = scalar;
	register Type	x = vector[0], y = vector[1], z = vector[2];

  // return L_row(q) in row-major order  (transpose result if col-major order)
  //
  L_q(0,0) =  s;  L_q(0,1) =  z;  L_q(0,2) = -y;  L_q(0,3) = -x;
  L_q(1,0) = -z;  L_q(1,1) =  s;  L_q(1,2) =  x;  L_q(1,3) = -y;
  L_q(2,0) =  y;  L_q(2,1) = -x;  L_q(2,2) =  s;  L_q(2,3) = -z;
  L_q(3,0) =  x;  L_q(3,1) =  y;  L_q(3,2) =  z;  L_q(3,3) =  s;

  return L_q;
}

template <typename Type>
Matrix<Type> Quat<Type>::R(void)
{
	Matrix<Type>	R_q(4,4);
	register Type	s = scalar;
	register Type	x = vector[0], y = vector[1], z = vector[2];

  // return R_row(q) in row-major order  (transpose result if col-major order)
  //
  R_q(0,0) =  s;  R_q(0,1) = -z;  R_q(0,2) =  y;  R_q(0,3) = -x;
  R_q(1,0) =  z;  R_q(1,1) =  s;  R_q(1,2) = -x;  R_q(1,3) = -y;
  R_q(2,0) = -y;  R_q(2,1) =  x;  R_q(2,2) =  s;  R_q(2,3) = -z;
  R_q(3,0) =  x;  R_q(3,1) =  y;  R_q(3,2) =  z;  R_q(3,3) =  s;

  return R_q;
}

template <typename Type>
Matrix<Type> Quat<Type>::quattomat(void)
{
	Quat<Type>	q(*this);

	Matrix<Type>	Q(4,4);
	register Type	s = scalar;
	register Type	x = vector[0], y = vector[1], z = vector[2];
	register Type	ss = s*s, sx = 2.0*s*x, sy = 2.0*s*y, sz = 2.0*s*z;
	register Type	          xx =     x*x, xy = 2.0*x*y, xz = 2.0*x*z;
	register Type	                        yy =     y*y, yz = 2.0*y*z;
	register Type	                                      zz =     z*z;

  // return (4x4) rot matrix, in row-major order we'd have:
  //   R(q^-1) . L(q) = Q
  // +-             -+ +-             -+
  // | s   z  -y   x | | s   z  -y  -x |
  // |-z   s   x   y | |-z   s   x  -y | =
  // | y  -x   s   z | | y  -x   s  -z |
  // |-x  -y  -z   s | | x   y   z   s |
  // +-             -+ +-             -+
  //
  // +-                                                                       -+
  // | (s^2+x^2-y^2-z^2)     (2xy+2sz)         (2xz-2sy)             0         |
  // |     (2xy-2sz)     (s^2-x^2+y^2-z^2)     (2yz+2sz)             0         |
  // |     (2xz+2sy)         (2yz-2sx)     (s^2-x^2-y^2+z^2)         0         |
  // |         0                 0                 0         (s^2+x^2+y^2+z^2) |
  // +-                                                                       -+
  //

  // both of these work, but the inverse and matrix mult involved are too
  // expensive
  //
  // return( (!q).R() * q.L() );
  // Q = (!q).R() * q.L();

  Q(0,0)=(ss+xx-yy-zz); Q(0,1)=(xy+sz); Q(0,2)=(xz-sy); Q(0,3)=0.0;
  Q(1,0)=(xy-sz); Q(1,1)=(ss-xx+yy-zz); Q(1,2)=(yz+sx); Q(1,3)=0.0;
  Q(2,0)=(xz+sy); Q(2,1)=(yz-sx); Q(2,2)=(ss-xx-yy+zz); Q(2,3)=0.0;
  Q(3,0)=0.0;     Q(3,1)=0.0;     Q(3,2)=0.0; Q(3,3)=(ss+xx+yy+zz);

  return(Q);
}

template <typename Type>
void Quat<Type>::mattoquat(Matrix<Type>& mat)
{
  // Routine to convert rotation matrix to a quaternion assuming matrix is 4x4
  // and quaternion is of unit magnitue.  The corresponding quaternion rotation 
  // matrix (R(q^-1) . L(q)) is given below:
  //
  // +-                                                    -+
  // | (1-2y^2-2z^2)       (2xy+2sz)        (2xz-2sy)     0 |
  // |    (2xy-2sz)     (1-2x^2-2z^2)       (2yz+2sz)     0 |
  // |    (2xz+2sy)        (2yz-2sx)     (1-2x^2-2y^2)    0 |
  // |        0                0                0         1 |
  // +-                                                    -+
  //
  // Note that if the quaternion is of unit magnitude, the diagonal entries
  // can be rewritten from the original (R(q^-1) . L(q)) form given above
  // in routine quattomat (i.e., the (0,0) entry (s^2+x^2-y^2-z^2) =
  // (1 - 2y^2 - 2z^2) iff s^2 + x^2 + y^2 + z^2 = 1).
  //
  // Code modified from:
  //   Watt, Alan and Watt, Mark, ``Advanced Animation and Rendering
  //   Techniques: Theory and Practice'', Addison-Wesley, Reading, MA, 1992.
  // (p.363)
  //

	Type	trace = mat.trace();
	Type	s=0.0;
	int	i,j,k,nxt[3] = {1,2,0};

  if( (mat.rows() != 4) || (mat.cols() != 4) )
    return;

  if(trace > 0.0) {
    s = sqrt(trace);
    scalar = s * 0.5;

    s *= 2.0;
    vector[0] = (mat(1,2) - mat(2,1))/s;
    vector[1] = (mat(2,0) - mat(0,2))/s;
    vector[2] = (mat(0,1) - mat(1,0))/s;
  }
  else {
    i = 0;
    if(mat(1,1) > mat(0,0)) i = 1;
    if(mat(2,2) > mat(i,i)) i = 2;
    j = nxt[i];  k = nxt[j];

    s = sqrt( (mat(i,i) - (mat(j,j) + mat(k,k))) + 1.0 );

    vector[i] = s * 0.5;		// s is one of x,y,z depending on i

    s *= 2.0;
    scalar = (mat(j,k) - mat(k,j))/s;
    vector[j] = (mat(i,j) + mat(j,i))/s;
    vector[k] = (mat(i,k) + mat(k,i))/s;
  }
}

template <typename Type>
Quat<Type> Quat<Type>::slerp2(Quat<Type> rhs, Type t)
{
  // Routine to calculate spherical linear interpolation (slerp) between
  // two quaternions (this and rhs) given paramter t.
  //
  // Code modified from:
  //   Watt, Alan and Watt, Mark, ``Advanced Animation and Rendering
  //   Techniques: Theory and Practice'', Addison-Wesley, Reading, MA, 1992.
  // (p.364)
  //
	Type		omega,sinom,cosom = dot(rhs);
	Type		sclp,sclq;
	Quat<Type>	p(this), q(rhs);
	Quat<Type>	qt(this);

  if((1.0+cosom) > EPSILON) {
    if((1.0-cosom) > EPSILON) {
      omega = acos((double)cosom);
      sinom = sin((double)omega);
      sclp = sin((1.0-t)*omega)/sinom;
      sclq = sin((    t)*omega)/sinom;
    }
    else {
      // if cos(omega) is very small, don't bother calculating sin(omega)
      // since sin(t) -> 0 as t -> 0.
      sclp = 1.0-t;
      sclq =     t;
    }
    qt = (sclp * p) + (sclq * q);
  }
  else {
    // Not quite sure what this piece of code does---see Watt & Watt.
    // I suspect this is the part that figures out whether we need to
    // calculate slerp(p,q,t) or slerp(p,-q,t), but it looks like it's
    // calculating slerp(p,-p,t) instead.  Possible bug in the book?
    // Since I'm a little leary about this, I added a type of API to slerp
    // which checks to see which arc we're traversing (see slerp()) based
    // on the distance of (p-q).(p-q) vs. (p+q).(p+q).  Due to that API routine
    // we may never actually get in here...
    //
    q.vector[0] = -p.vector[1];
    q.vector[1] =  p.vector[0];
    q.vector[2] = -p.scalar;

    sclp = sin((1.0-t)*M_PI_2);
    sclq = sin((    t)*M_PI_2);
    qt = (sclp * p) + (sclq * q);

    qt.scalar   =  p.vector[2];
  }

  return(qt);
}

template <typename Type>
Quat<Type> Quat<Type>::slerp(Quat<Type> rhs, Type t)
{
  // Routine to calculate spherical linear interpolation (slerp) between
  // two quaternions (this and rhs) given paramter t.  This is sort of an
  // API to slerp2 which does most of the work.  Here, I put in the test
  // to see whether we need to negate q (rhs) or not as per the text:
  //   Watt, Alan and Watt, Mark, ``Advanced Animation and Rendering
  //   Techniques: Theory and Practice'', Addison-Wesley, Reading, MA, 1992.
  // (p.365)
  //
	Quat<Type>	p(this),q(rhs);
	Quat<Type>	p_pos_q_diff(p - q),p_neg_q_diff(p + q);

  if(p_pos_q_diff.sqlength() < p_neg_q_diff.sqlength())
    return(slerp2( rhs,t));
  else
    return(slerp2(-rhs,t));
}

/////////////////////////// protected helper functions (for constructors) //////

template <typename Type>
void Quat<Type>::trackball(Type x1,Type y1,Type x2,Type y2)
{
  // Routine to simulate a trackball, modified from code originally written by
  // Gavin Bell for Silicon Graphics, November 1988.  His (presumably) comments 
  // follow.  (My comments appear in <>'s.)
  //
  // Ok, simulate a track-ball.  Project the points onto the virtual trackball,
  // then figure out the axis of rotation, which is the cross product of P1 P2
  // and O P1 (O is the center of the ball, 0,0,0).
  //
  // Note: This is a deformed trackball--is a trackball in the center, but is
  // deformed into a hyperbolic sheet of rotation away from the center.  This
  // particular function was chosen after trying out several variations.
  //
  // It is assumed that the arguments to this routine are in the range
  // (-1.0 ... 1.0)
  //

	Type		rangle;
	Vector<Type>	*p1,*p2,*d,*axis;

  if( (x1 == x2) && (y1 == y2) ){
    // zero rotation
    //
    // <note: this routine is intended as part of a constructor, so we really
    // don't need this stuff below, since vector and scalar have already been
    // initialized to these values---I left the lines in though just in case
    // I ever decide to make this a publicly accessible routine.>
    //
    vector[0] = 0.0;
    vector[1] = 0.0;
    vector[2] = 0.0;
    scalar = 1.0;
    return;
  }

  // First, figure out z-coordinates for projection of P1 and P2 to deformed
  // sphere.
  p1 = new Vector<Type>(x1,y1,project_to_sphere(TRACKBALLSIZE,x1,y1));
  p2 = new Vector<Type>(x2,y2,project_to_sphere(TRACKBALLSIZE,x2,y2));

  // Now, we want the cross product of P1 and P2.
  axis = new Vector<Type>(p2->cross(p1));

  // Figure out how much to rotate around that axis.
  d = new Vector<Type>((*p1) - (*p2));
  rangle = d->length() / (2.0 * TRACKBALLSIZE);

  // Avoid problems with out-of-control values...<assume rangle is in radians>
  if(rangle >  1.0) rangle =  1.0;
  if(rangle < -1.0) rangle = -1.0;
  rangle = (Type)(2.0 * asin((double)rangle));

  // <Set this quaternion's s and v values.>
  scalar = cos(rangle/2.0);
  vector = axis->norm(); // <normalize axis of rotation (e.g., use unit vector)>
  vector *= sin(rangle/2.0);

  // <Collect garbage.>
  delete p1; delete p2; delete d; delete axis;
}

template <typename Type>
Type Quat<Type>::project_to_sphere(Type r,Type x,Type y)
{
  // Helper routine for trackball simulation, modified from code originally
  // written by // Gavin Bell for Silicon Graphics, November 1988.  His
  // (presumably) comments follow.  (My comments, if any, appear in <>'s.)
  //
  // Project an x,y pair onto a sphere of radius r OR a hyperbolic sheet if
  // we are away from the center of the sphere.
  //
	Type	d,t,z;

  d = sqrt(x*x + y*y);
  if(d < r * ROOT2OVER2)	// inside sphere
    z = sqrt(r*r - d*d);
  else {			// on hyperbola
    t = r / ROOT2;
    z = t*t / d;
  }
  return z;
}
