#ifndef QUAT_H
#define QUAT_H

#include	"matrix.h"
#include	"vector.h"

#define	TRACKBALLSIZE	(0.8)
#define	ROOT2		(1.41421356237309504880)
#define	ROOT2OVER2	(.70710678118654752440)
#define	EPSILON		(0.00001)

template <typename Type>
class Quat {
	public:
	// constructors (overloaded)
	Quat(Type sin=0.0) : \
		scalar(sin), \
		vector(3)			{ };
	Quat(const Quat& rhs) : \
		scalar(rhs.scalar), \
		vector(rhs.vector)		{ };
	Quat(Quat *rhs) : \
		scalar(rhs->scalar), \
		vector(rhs->vector)		{ };
	Quat(Type sin,Vector<Type> vin) : \
		scalar(sin), \
		vector(vin)			{ };
	Quat(Matrix<Type>& m) : \
		scalar(1.0), \
		vector(0.0,0.0,0.0)		{ mattoquat(m); }
	Quat(Type x1,Type y1,Type x2, Type y2) : \
		scalar(1.0), \
		vector(0.0,0.0,0.0)		{ trackball(x1,y1,x2,y2); }

	// destructors
	//~Quat()					{ delete &vector; }

	// operators: unary
	Quat operator-(void);
	Quat operator!(void);			// quaternion inverse

	// operators: (Quat, const Matrix&)
        Quat operator*(const Matrix<Type>& rhs);

	// operators: (Quat, const Quat&)
        int operator==(const Quat& rhs);
        int operator<(const Quat& rhs);
        int operator>(const Quat& rhs);
        Quat operator=(const Quat& rhs);
        Quat operator*=(const Quat& rhs);
        Quat operator*(const Quat& rhs);
        Quat operator/=(Quat& rhs);
        Quat operator/(Quat& rhs);
        Quat operator+=(const Quat& rhs);
        Quat operator+(const Quat& rhs);
        Quat operator-=(const Quat& rhs);
        Quat operator-(const Quat& rhs);

	// operators: (Quat, Quat*)
        int operator==(Quat* rhs)		{ return((*this) == (*rhs)); }
        int operator<(Quat* rhs)		{ return((*this) == (*rhs)); }
        int operator>(Quat* rhs)		{ return((*this) == (*rhs)); }
        Quat operator=(Quat* rhs)		{ return((*this)  = (*rhs)); }
        Quat operator*=(Quat* rhs)		{ return((*this) *= (*rhs)); }
        Quat operator*(Quat* rhs)		{ return((*this) *  (*rhs)); }
        Quat operator/=(Quat* rhs)		{ return((*this) /= (*rhs)); }
        Quat operator/(Quat* rhs)		{ return((*this) /  (*rhs)); }
        Quat operator+=(Quat* rhs)		{ return((*this) += (*rhs)); }
        Quat operator+(Quat* rhs)		{ return((*this) +  (*rhs)); }
        Quat operator-=(Quat* rhs)		{ return((*this) -= (*rhs)); }
        Quat operator-(Quat* rhs)		{ return((*this) -  (*rhs)); }

	// operators (const Quat&, scalar)
        Quat operator*=(const Type& rhs);
        Quat operator*(const Type& rhs);
        Quat operator/=(const Type& rhs);
        Quat operator/(const Type& rhs);

	// friend operators: (const scalar*, const Quat&)
	friend Quat operator*(Type lhs,Quat& rhs);
	friend Quat operator/(Type lhs,Quat& rhs);

	// friends
	friend ostream& operator<<(ostream& s, const Quat& rhs);
//	friend ostream& operator<<(ostream& s, Quat *rhs)
//	  { return(s << (*rhs)); }

	// member functions
	Type&		s()			{ return(scalar); }
	Vector<Type>&	v()			{ return(vector); }
	Type		dot(Quat& rhs);
	Type		dot(Quat* rhs)		{ return(dot((*rhs))); }
	Type		sqlength(void)		{ return(dot(this)); }
	Type		length(void)		{ return(sqrt(dot(this))); }
	Quat		norm(void)		{ return(*this / length()); }
	Matrix<Type>	L(void);
	Matrix<Type>	R(void);
	Matrix<Type>	quattomat(void);	// return (4x4) rot matrix
	void		mattoquat(Matrix<Type>& m);
	Quat		slerp2(Quat rhs,Type t);// spherical linear interp.
	Quat		slerp(Quat rhs,Type t);// spherical linear interp.

	// protected: available to this (base) class and subs. derived classes
	protected:
	void		trackball(Type x1,Type y1,Type x2,Type y2);
	Type		project_to_sphere(Type r,Type x,Type y);

	// private: only available to this class
	private:
	Type		scalar;		// scalar part
	Vector<Type>	vector;		// real components of imaginary part
};

#endif
