/****************************************************************************
** $Id: qt/gltexobj.h   3.1.2   edited Nov 8 2002 $
**
** Copyright (C) 1992-2002 Trolltech AS.  All rights reserved.
**
** This file is part of an example program for Qt.  This example
** program may be used, distributed and modified without limitation.
**
*****************************************************************************/

/****************************************************************************
**
** This is a simple QGLWidget displaying an openGL wireframe box
**
****************************************************************************/

#ifndef GLTEXOBJ_H
#define GLTEXOBJ_H

#include <Cg/cg.h>
#include <Cg/cgGL.h>
#include <qgl.h>

using namespace std;

#include <mlib/types.h>
#include <mlib/array.h>
#include <mlib/matrix.h>
#include <mlib/vector.h>
#include <mlib/quaternion.h>

#include <imglib/lum.h>
#include <imglib/rgba.h>
#include <imglib/smpte.h>
#include <imglib/pnm.h>
#include <imglib/ppm.h>
#include <imglib/pgm.h>

#include <timerlib/timer.h>

typedef enum { NVIDIA, ATI } vendor_t;

class GLTexobj : public QGLWidget
{
	Q_OBJECT

public:

	GLTexobj( QWidget* parent, const char* name );
	~GLTexobj();

public slots:

	void		mouseMoveEvent(QMouseEvent* e);
	void		update();
	void		konnect() {};
	void		calibrate() {};
	void		imageOpen();
	void		imageTexture();
	void		imageMask();
	void		imageLOD(int);
	void		imageDeg(int);
        void            openColorDeg();
        void            openSpatDeg();
	void		trackRun() {};
	void		trackStop() {};
	void		trackStatus() {};

protected:

	void		initializeGL();
	void		paintGL();
	void		resizeGL(int w, int h);

	void		chooseCgProfiles();
	void		loadCgPrograms();
	void		loadCgTextures();

private:
	GLuint		texNames[2];
	QTimer*		timer;
	float		dx,dy;
	Ppm		*img, *mask;
	vendor_t	vendor;
	double		fps;
	double		fps_samples;
	double		fps_avg;

	int		lod;

	CGcontext	context;
	CGprogram	vertexProgram;
	CGprogram	fragmentProgram;
	CGprofile	vertexProfile;
	CGprofile	fragmentProfile;
	bool		haveLoadedPrograms;
};


#endif // GLTEXOBJ_H
